Type: Package
Package: rex
Title: Friendly Regular Expressions
Version: 1.2.2
Authors@R: c(
    person("Kevin", "Ushey", , "kevin@rstudio.com", role = c("aut", "cre")),
    person("Jim", "Hester", , "james.f.hester@gmail.com", role = "aut"),
    person("Robert", "Krzyzanowski", , "rkrzyzanowski@gmail.com", role = "aut")
  )
Description: A friendly interface for the construction of regular
    expressions.
License: MIT + file LICENSE
URL: https://rex.r-lib.org, https://github.com/r-lib/rex
BugReports: https://github.com/r-lib/rex/issues
Suggests: covr, dplyr, ggplot2, Hmisc, knitr, magrittr, rmarkdown,
        roxygen2, rvest, stringr, testthat
VignetteBuilder: knitr
Encoding: UTF-8
RoxygenNote: 7.1.2
Collate: 'aaa.R' 'utils.R' 'escape.R' 'capture.R' 'character_class.R'
        'counts.R' 'lookarounds.R' 'match.R' 'or.R' 'rex-mode.R'
        'rex.R' 'shortcuts.R' 'wildcards.R' 'zzz.R'
Config/Needs/website: r-lib/pkgdown, tidyverse/tidytemplate
Imports: withr
NeedsCompilation: no
Packaged: 2026-03-27 21:22:29 UTC; kevin
Author: Kevin Ushey [aut, cre],
  Jim Hester [aut],
  Robert Krzyzanowski [aut]
Maintainer: Kevin Ushey <kevin@rstudio.com>
Repository: CRAN
Date/Publication: 2026-03-28 16:10:02 UTC
Built: R 4.6.0; ; 2026-04-26 09:42:28 UTC; unix

# Generated by roxygen2: do not edit by hand

S3method(as.character,regex)
S3method(as.regex,default)
S3method(c,regex)
S3method(character_class_escape,character)
S3method(character_class_escape,character_class)
S3method(character_class_escape,default)
S3method(character_class_escape,list)
S3method(character_class_escape,regex)
S3method(escape,character)
S3method(escape,character_class)
S3method(escape,default)
S3method(escape,list)
S3method(escape,regex)
S3method(print,regex)
export(as.regex)
export(character_class)
export(character_class_escape)
export(escape)
export(m)
export(matches)
export(re_matches)
export(re_substitutes)
export(regex)
export(register_shortcuts)
export(rex)
export(rex_)
export(rex_mode)
export(s)
export(shortcuts)
export(substitutes)

This archive contains the distribution Rex,
version 1.12.0:

  the friendly automation framework

This software is Copyright (c) 2020 by Jan Gehring.

This is free software, licensed under:

  The Apache License, Version 2.0, January 2004


This README file was generated by Dist::Zilla::Plugin::Readme v6.015.

# Rex

<!-- badges: start -->
[![Codecov test coverage](https://codecov.io/gh/kevinushey/rex/branch/master/graph/badge.svg)](https://app.codecov.io/gh/kevinushey/rex?branch=main)
[![Lifecycle: stable](https://img.shields.io/badge/lifecycle-stable-brightgreen.svg)](https://lifecycle.r-lib.org/articles/stages.html)
[![R-CMD-check](https://github.com/r-lib/rex/actions/workflows/R-CMD-check.yaml/badge.svg)](https://github.com/r-lib/rex/actions/workflows/R-CMD-check.yaml)
<!-- badges: end -->

### Friendly Regular Expressions

Regular expressions are very powerful feature, however they are often difficult
to interpret. Rex allows you to build complex regular expressions from human
readable expressions.  So instead of writing (and later trying to decipher)
```r
r <- "^(?:(((?:[^:])+)://))?((?:(?:(?!:/).)*)+)(?:(:([[:digit:]]+)))?(?:(/.*))?$"
```

You can write

```r
r <- rex(

  start,

  ## match the protocol -- may exist or may not
  maybe(capture(
      capture(except_some_of(":")),
      "://"
      )),

  ## match the path
  capture(one_or_more(not(":/"))),

  ## get the port
  maybe(capture(":", capture(numbers))),

  ## and the rest
  maybe(capture("/", anything)),

  end

)
```

While these expressions are a bit longer than their corresponding regular
expression, they are much more readable and maintainable.

## Installation

```r
install.packages("rex")
```

## Usage

The vignettes have longer form usage examples.

- [URL Validation](http://rpubs.com/jimhester/rex-url_parsing)
- [Webserver Log Parsing](http://rpubs.com/jimhester/rex-log_parsing)

Each `rex()` function call can include a number of functions and shortcuts.
For a full list of the functions available please see `?rex` and `?shortcuts`.

### Rex Mode

Rex functions are not exported because they are only useful within `rex()`
calls, but they can be temporarily attached using `rex_mode()` which allows
them to be auto-completed.

### Using Rex in other packages

Using `rex` in other packages will generate spurious NOTEs from `R CMD check`
unless you include a call to `rex::register_shortcuts()` with your package name
somewhere in your package source.  This function registers all of the rex
shortcuts as valid variables fixing the NOTEs.

## See Also
- [Regularity](https://github.com/andrewberls/regularity) - Ruby library that
  partially inspired `rex`.
- [PCRE](https://www.pcre.org/) - Perl Compatible Regular Expressions, the
  engine that `rex` regular expressions use.
- [Perl 5 Regular Expressions](https://perldoc.perl.org/perlre) - Perl
  regular expression documentation, which are nearly 100% compatible with PCRE.

# 🦖 Rex

[![license](https://img.shields.io/badge/license-MIT-blue.svg)](LICENSE)
[![docs](https://img.shields.io/badge/docs-online-blue)](https://docs.rs/rex)
[![crates.io](https://img.shields.io/crates/v/rex.svg)](https://crates.io/crates/rex)

<p align="center">
  <img src="logo.jpg" width="400">
</p>

Rex (short for *[Rush](https://rush.cloud/) Expressions*) is a strongly-typed,
pure, implicitly parallel functional programming language built to be an
excellent target for LLM-generated programs, with a focus on scientific
workflows and data analysis pipelines. At a high level, you write
transformations over lists, records, ADTs, and other values using familiar
functional building blocks like `map`, `filter`, folds, pattern matching, and
composition. The language is designed to make dataflow clear and predictable,
with types and pure expressions doing most of the heavy lifting.

Rex is designed first and foremost to be embedded inside Rust applications. In
that model, your Rust program acts as the host runtime and injects native
functions into Rex so scripts can orchestrate real work while staying in a
concise, declarative style. This makes Rex a practical scripting layer for
workflow-style systems where you want strong typing and explicit control at the
host boundary.

Because Rex programs are pure and free of side effects in the language itself,
the runtime can safely execute host-provided async functions in parallel when
it is valid to do so. In practice, that means users can write straightforward
functional code and still benefit from concurrency without directly managing
threads, locks, or low-level async orchestration.

If you are using Rex as a code-generation target, read
**[LLM Guidance](docs/src/LLMS.md)** early. It captures syntax pitfalls and
validation workflow that reduce iteration time.

## Example

```rex
let
  values = [3, 12, 7, 20, 15, 4],
  selected = filter (\n -> n >= 10) values,
  adjusted = map (\n -> n - 2) selected,
  total = foldl (\acc n -> acc + n) 0 adjusted
in
  (values, selected, adjusted, total)
```

[Try it yourself in the interactive browser-based playground](https://peterkelly.github.io/rex/)

## Rex as a target for LLMs

<p align="center">
  <img src="agents_rex.jpg" width="600">
</p>

Rex is the world’s first parallel functional language explicitly designed to be
a useful target for LLMs. Its strong static type system gives rapid,
high-signal feedback on generated programs, so both users and models can
quickly identify mismatches and converge on correct code.

That typechecking loop works especially well with Rex’s functional,
expression-oriented style. Because programs are written as pure data
transformations, LLM-generated code tends to be easier to inspect, reason
about, and refine than imperative scripts with hidden state or side effects.

Together, these properties make Rex a strong fit for LLM-generated data
analysis pipelines and scientific workflows. Models can generate high-level
orchestration in Rex, while host-provided Rust functions handle domain-specific
execution, giving a clean split between deterministic workflow logic and host
capabilities.

## Documentation

[https://peterkelly.github.io/rex/](https://peterkelly.github.io/rex/)

## Current status

The `main` branch of this repository contains what will soon become Rex v4.
Currently it is marked as 3.9.x until the API has stabilized, at which point
the version number will be updated.

The most recent stable (and production-ready) version is available at
https://github.com/talo/rex.

## Crates

This repo is a Cargo workspace. The key crates are:

- `rex-parser`: source parser producing a `CompilationUnit { decls, body }`
- `rex-ast`: AST nodes, symbols, and spans
- `rex-typesystem`: Hindley–Milner type inference + type classes + ADTs
- `rex-engine`: runtime evaluator + native-function injection, backed by `rex-typesystem`
- `rex-proc-macro`: `#[derive(Rex)]` for bridging Rust types ↔ Rex ADTs/values
- `rex-cli`: CLI crate providing the `rex` binary (`cargo run -p rex-cli --bin rex -- ...`)
- `rex-fuzz`: stdin-driven fuzz harness binaries
- `rex-util`: small shared helpers (e.g. module hashing, bundled `std.*` module sources)
- `rex-lsp` / `rex-vscode`: language tooling (LSP + VS Code extension)

## CLI

Run a snippet file:

```sh
cargo run -p rex-cli --bin rex -- --snippet rex-cli/examples/record_update.rex
```

Run the advanced module import example:

```sh
cargo run -p rex-cli --bin rex -- --snippet rex-cli/examples/modules_advanced/main.rex
```

Run inline code:

```sh
cargo run -p rex-cli --bin rex -- -c 'map ((*) 2) [1, 2, 3]'
```

Other useful flags:

- `--emit-ast`: print parsed AST as JSON and exit
- `--emit-type` (alias: `--type`): print inferred type as JSON and exit
- `--stdin`: read a program from stdin
- `--stack-size-mb`: control the runner thread stack size

## Standard Module (Prelude)

Rex ships with a prelude that provides core functions, types, and type classes
used throughout the language (for example mapping, filtering, folds, numeric
operations, equality/ordering, and container abstractions).

The prelude surface definitions and type-class wiring live primarily in:

- `rex-typesystem/src/prelude_typeclasses.rex`
- `rex-typesystem/src/prelude.rs`
- `rex-engine/src/prelude.rs`

For full details and usage patterns, see the [docs](https://peterkelly.github.io/rex/):

- [Prelude tour](https://peterkelly.github.io/rex/tutorial/section1/12_prelude_tour.html)
- [Language reference](https://peterkelly.github.io/rex/LANGUAGE.html)
- [LLM usage guidelines](https://peterkelly.github.io/rex/LLMS.html)

Made with ❤️ by [QDX](https://qdx.co/)

## ----include = FALSE----------------------------------------------------------
library(rex)
library(dplyr)
library(knitr)
library(ggplot2)
library(magrittr)

## ----show.warnings=FALSE------------------------------------------------------
parsed <- scan("NASA.txt", what = "character", sep = "\n") %>%
  re_matches(
    rex(

      # Get the time of the request
      "[",
        capture(name = "time",
          except_any_of("]")
        ),
      "]",

      space, double_quote, "GET", space,

      # Get the filetype of the request if requesting a file
      maybe(
        non_spaces, ".",
        capture(name = "filetype",
          except_some_of(space, ".", "?", double_quote)
        )
      )
    )
  ) %>%
  mutate(filetype = tolower(filetype),
         time = as.POSIXct(time, format="%d/%b/%Y:%H:%M:%S %z"))

## ----echo = FALSE-------------------------------------------------------------
kable(head(parsed, n = 10))

## ----FALSE, fig.show='hold', warning = FALSE, message = FALSE-----------------
ggplot(na.omit(parsed)) + stat_count(aes(x=filetype))
ggplot(na.omit(parsed)) + geom_histogram(aes(x=time)) + ggtitle("Requests over time")

## ----url_parsing_stock, eval=F------------------------------------------------
# "^(?:(?:http(?:s)?|ftp)://)(?:\\S+(?::(?:\\S)*)?@)?(?:(?:[a-z0-9\u00a1-\uffff](?:-)*)*(?:[a-z0-9\u00a1-\uffff])+)(?:\\.(?:[a-z0-9\u00a1-\uffff](?:-)*)*(?:[a-z0-9\u00a1-\uffff])+)*(?:\\.(?:[a-z0-9\u00a1-\uffff]){2,})(?::(?:\\d){2,5})?(?:/(?:\\S)*)?$"

## ----url_parsing_url----------------------------------------------------------
library(rex)
library(magrittr)

valid_chars <- rex(except_some_of(".", "/", " ", "-"))

re <- rex(
  start,

  # protocol identifier (optional) + //
  group(list("http", maybe("s")) %or% "ftp", "://"),

  # user:pass authentication (optional)
  maybe(non_spaces,
    maybe(":", zero_or_more(non_space)),
    "@"),

  #host name
  group(zero_or_more(valid_chars, zero_or_more("-")), one_or_more(valid_chars)),

  #domain name
  zero_or_more(".", zero_or_more(valid_chars, zero_or_more("-")), one_or_more(valid_chars)),

  #TLD identifier
  group(".", valid_chars %>% at_least(2)),

  # server port number (optional)
  maybe(":", digit %>% between(2, 5)),

  # resource path (optional)
  maybe("/", non_space %>% zero_or_more()),

  end
)

## ----url_parsing_validate-----------------------------------------------------
good <- c("http://foo.com/blah_blah",
  "http://foo.com/blah_blah/",
  "http://foo.com/blah_blah_(wikipedia)",
  "http://foo.com/blah_blah_(wikipedia)_(again)",
  "http://www.example.com/wpstyle/?p=364",
  "https://www.example.com/foo/?bar=baz&inga=42&quux",
  "http://✪df.ws/123",
  "http://userid:password@example.com:8080",
  "http://userid:password@example.com:8080/",
  "http://userid@example.com",
  "http://userid@example.com/",
  "http://userid@example.com:8080",
  "http://userid@example.com:8080/",
  "http://userid:password@example.com",
  "http://userid:password@example.com/",
  "http://➡.ws/䨹",
  "http://⌘.ws",
  "http://⌘.ws/",
  "http://foo.com/blah_(wikipedia)#cite-1",
  "http://foo.com/blah_(wikipedia)_blah#cite-1",
  "http://foo.com/unicode_(✪)_in_parens",
  "http://foo.com/(something)?after=parens",
  "http://☺.damowmow.com/",
  "http://code.google.com/events/#&product=browser",
  "http://j.mp",
  "ftp://foo.bar/baz",
  "http://foo.bar/?q=Test%20URL-encoded%20stuff",
  "http://مثال.إختبار",
  "http://例子.测试",
  "http://-.~_!$&'()*+,;=:%40:80%2f::::::@example.com",
  "http://1337.net",
  "http://a.b-c.de",
  "http://223.255.255.254")

bad <- c(
  "http://",
  "http://.",
  "http://..",
  "http://../",
  "http://?",
  "http://??",
  "http://??/",
  "http://#",
  "http://##",
  "http://##/",
  "http://foo.bar?q=Spaces should be encoded",
  "//",
  "//a",
  "///a",
  "///",
  "http:///a",
  "foo.com",
  "rdar://1234",
  "h://test",
  "http:// shouldfail.com",
  ":// should fail",
  "http://foo.bar/foo(bar)baz quux",
  "ftps://foo.bar/",
  "http://-error-.invalid/",
  "http://-a.b.co",
  "http://a.b-.co",
  "http://0.0.0.0",
  "http://3628126748",
  "http://.www.foo.bar/",
  "http://www.foo.bar./",
  "http://.www.foo.bar./")

all(grepl(re, good) == TRUE)

all(grepl(re, bad) == FALSE)

#' @include utils.R
NULL

#' Escape characters for a regex
#'
#' @param x Object to escape.
#' @export
escape <- function(x) UseMethod("escape")

#' @describeIn escape Objects are simply passed through unchanged.
#' @export
escape.regex <- function(x) x

#' @describeIn escape Objects are surrounded by braces.
#' @export
escape.character_class <- function(x) {
  p("[", x, "]")
}

#' @describeIn escape Objects are properly escaped for regular expressions.
#' @export
escape.character <- function(x) {
  chars <- c("*", ".", "?", "^", "+", "$", "|", "(", ")", "[", "]", "{", "}", "\\")
  regex(sanitize(x, chars))
}

#' @describeIn escape default escape coerces to character and escapes.
#' @export
escape.default <- function(x) {
  escape.character(as.character(x))
}

#' @describeIn escape simply call escape on all elements of the list.
#' @export
escape.list <- function(x) {
  lapply(x, escape)
}

escape_dots <- function(...) {
  unlist(escape(eval(list(...))))
}

#' Character class escapes
#' @inheritParams escape
#' @export
character_class_escape <- function(x) UseMethod("character_class_escape")

#' @describeIn character_class_escape objects are passed through unchanged.
#' @export
character_class_escape.regex <- function(x) x

#' @describeIn character_class_escape objects are passed through unchanged.
#' @export
character_class_escape.character_class <- character_class_escape.regex

#' @describeIn character_class_escape objects properly escaped for character classes.
#' @export
character_class_escape.character <- function(x) {
  regex(sanitize(x, c("-", "^", "[", "]", "\\")))
}

#' @describeIn character_class_escape call \code{character_class_escape} on all elements of the list.
#' @export
character_class_escape.list <- function(x) {
  lapply(x, character_class_escape)
}

#' @describeIn character_class_escape coerce to \code{character} and \code{character_class_escape}.
#' @export
character_class_escape.default <- function(x) {
  character_class_escape.character(as.character(x))
}

character_class_escape_dots <- function(...) {
  unlist(character_class_escape(eval(list(...))))
}

sanitize <- function(x, chars) {
  gsub(paste0("([\\", paste0(collapse = "\\", chars), "])"), "\\\\\\1", x, perl = TRUE)
}

.onAttach <- function(lib, pkg) { # nolint
  if (!interactive() || stats::runif(1) > 0.1) return()

  packageStartupMessage("Welcome to rex, the friendly regular expression helper!\n",
                        "Use 'rex_mode()' to toggle code completion for rex shortcuts and functions.")
}

#' @include escape.R
#' @include utils.R
NULL

#' Create a capture group
#'
#' Used to save the matched value within the group for use later in the regular
#' expression or to extract the values captured.  Both named and unnamed groups
#' can later be referenced using \code{\link{capture_group}}.
#'
#' @param name of the group.  Unnamed capture groups are numbers starting at 1
#' in the order they appear in the regular expression.  If two groups have the
#' same name, the leftmost group is the used in any reference.
#' @param ... \code{\link{shortcuts}}, R variables, text, or other \pkg{rex}
#' functions.
#' @family rex
#' @aliases .
#' @seealso \code{\link{group}} for grouping without capturing.  Perl 5 Capture
#' Groups \url{http://perldoc.perl.org/perlre.html#Capture-groups}
#' @examples
#'
#' # Match paired quotation marks
#' re <- rex(
#'   # first quotation mark
#'   capture(quotes),
#'
#'   # match all non-matching quotation marks
#'   zero_or_more(except(capture_group(1))),
#'
#'   # end quotation mark (matches first)
#'   capture_group(1)
#' )
#'
#' #named capture - don't match apples to oranges
#' re <- rex(
#'   capture(name = "fruit", or("apple", "orange")),
#'   "=",
#'   capture_group("fruit")
#' )
capture <- . <- function(..., name = NULL) {
  if(!is.null(name)) {
    name <- paste0("?<", name, ">")
  }
  p( "(", name, p(escape_dots(...)), ")" )
}
register(capture, .)

#' @rdname capture
capture_group <- function(name) {
  p( "\\g{", name, "}" )
}
register(capture_group)

#' Create a grouped expression
#'
#' This is similar to \code{\link{capture}} except that it does not store the
#' value of the group.  Best used when you want to combine several parts
#' together and do not reference or extract the grouped value later.
#' @inheritParams capture
#' @seealso \code{\link{capture}} for grouping with capturing.  Perl 5 Extended
#' Patterns \url{http://perldoc.perl.org/perlre.html#Extended-Patterns}
#' @family rex
group <- function(...) {
  p( "(?:", p(escape_dots(...)), ")" )
}
register(group)

#' @include escape.R
#' @include utils.R
NULL

#' Counts
#'
#' Functions to restrict a regex to a specific number
#' @param x A regex pattern.
#' @param n An integer number
#' @param low An integer number for the lower limit.
#' @param high An integer number for the upper limit.
#' @inheritParams zero_or_more
#' @family rex
#' @name counts
NULL

#' @aliases n
#' @describeIn counts \code{x} must occur exactly \code{n} times.
n_times <- n <- function(x, n, type = c("greedy", "lazy", "possessive")) {
  add_type(p("(?:", p(escape(x)), "){", n, "}"), type)
}
register(n_times, n)

#' @describeIn counts \code{x} must occur between \code{low} and \code{high} times.
between <- function(x, low, high, type = c("greedy", "lazy", "possessive")) {
  add_type(p("(?:", p(escape(x)), "){", low, ",", high, "}"), type)
}
register(between)

#' @describeIn counts \code{x} must occur at least \code{n} times.
at_least <- function(x, n, type = c("greedy", "lazy", "possessive")) {
  add_type(between(x, n, ""), type)
}
register(at_least)

#' @describeIn counts \code{x} must occur at most \code{n} times.
at_most <- function(x, n, type = c("greedy", "lazy", "possessive")) {
  add_type(between(x, "", n), type)
}
register(at_most)

context("n_times")
`%>%` <- magrittr::`%>%`
test_that("description", {
  re1 <- rex("x" %>% n_times(2))
  re2 <- rex("x" %>% n(2))

  expect_identical(re1, re2)

  expect_equal(re1,
    regex("(?:x){2}"))

  expect_true(grepl(re1, "xx"))
  expect_false(grepl(re1, "x"))
})

context("between")
test_that("creates a bounded repetition", {
  re <- rex("x" %>% between(2, 4))
  expect_equal(re, regex("(?:x){2,4}"))

  expect_true(grepl(re, "xxx"))
  expect_false(grepl(re, "x"))
})

context("at_least")
test_that("creates a bounded repetition", {
  re <- rex("x" %>% at_least(3))
  expect_equal(re, regex("(?:x){3,}"))

  expect_true(grepl(re, "xxx"))
  expect_false(grepl(re, "xx"))
})

context("at_most")
test_that("creates a repetition of n times at most", {
  re <- rex(start, "x" %>% at_most(3), end)
  expect_equal(re, regex("^(?:x){,3}$"))

  expect_true(grepl(re, "xxx"))
  expect_false(grepl(re, "xxxxx"))
})

#' Match function
#'
#' @param data character vector to match against
#' @param pattern regular expression to use for matching
#' @param global use global matching
#' @param options regular expression options
#' @param locations rather than returning the values of the matched (or
#' captured) string, return a \code{data.frame} of the match locations in the
#' string.
#' @param ... options passed to regexpr or gregexpr
#' @return if no captures, returns a logical vector the same length as the
#' input character vector specifying if the relevant value matched or not.  If
#' there are captures in the regular expression, returns a \code{data.frame} with a
#' column for each capture group.  If \code{global} is \code{TRUE}, returns a
#' list of \code{data.frame}s.
#' @seealso \code{\link{regexp}} Section "Perl-like Regular Expressions" for a
#' discussion of the supported options
#' @examples
#' string <- c("this is a", "test string")
#' re_matches(string, rex("test")) # FALSE FALSE
#'
#' # named capture
#' re_matches(string, rex(capture(alphas, name = "first_word"), space,
#'   capture(alphas, name = "second_word")))
#' #   first_word second_word
#' # 1       this          is
#' # 2       test      string
#'
#' # capture returns NA when it fails to match
#' re_matches(string, rex(capture("test")))
#' #      1
#' # 1 test
#' # 2 <NA>
#' @aliases matches m
#' @export re_matches matches m
re_matches <- matches <- m <- function(data, pattern, global = FALSE, options = NULL, locations = FALSE, ...) {

  pattern <- add_options(pattern, options)

  process_matches <- function(match, string) {

    if(no_capture(match)) {

      # if no capture and no location just return if the regex matched
      if(!locations) {
        return(match != -1L)
      }

      # else return a data frame of the start and end locations
      match[ match == -1L ] <- NA_integer_
      starts <- match
      attributes(starts) <- NULL

      lengths <- attr(match, "match.length")
      ends <- starts + lengths - 1L

      return(data.frame(start = starts, end = ends))
    }

    # if a capture return a data frame with the capture results for each string
    starts <- attr(match, "capture.start")
    lengths <- attr(match, "capture.length")
    ends <- starts + lengths - 1L

    not_matched <- starts == -1L

    strings <- substring(string, starts, ends)

    strings[not_matched] <- NA_character_

    res <- matrix(ncol = ncol(starts), strings)

    nms <- auto_name(attr(match, "capture.names"))

    if (!locations) {
      colnames(res) <- nms
      return(as.data.frame(res, stringsAsFactors = FALSE, check.names = FALSE))
    }

    starts[not_matched] <- NA_integer_

    ends[not_matched] <- NA_integer_

    indexes <- unlist(lapply(seq_len(ncol(res)), function(x) {
        seq(x, by = ncol(res), length.out = 3)
    }))

    full <- data.frame(res, starts, ends, stringsAsFactors = FALSE, check.names = FALSE)[, indexes, drop = FALSE]
    full_names <- unlist(Map(function(name) c(name, paste(sep=".", name, c("start", "end"))), nms, USE.NAMES=F))
    colnames(full) <- full_names

    full
  }

  if(global %==% TRUE) {
    mapply(process_matches, gregexpr(pattern = pattern, data, perl = TRUE, ...), data, SIMPLIFY = FALSE)
  }
  else {
    process_matches(regexpr(pattern = pattern, data, perl = TRUE, ...), data)
  }
}

#' Substitute regular expressions in a string with another string.
#'
#' @param data character vector to substitute
#' @param pattern regular expression to match
#' @param replacement replacement text to use
#' @param global substitute all occurrences
#' @param options option flags
#' @param ... options passed to sub or gsub
#' @seealso \code{\link{regexp}} Section "Perl-like Regular Expressions" for a
#' discussion of the supported options
#' @examples
#' string <- c("this is a Test", "string")
#' re_substitutes(string, "test", "not a test", options = "insensitive")
#' re_substitutes(string, "i", "x", global = TRUE)
#' re_substitutes(string, "(test)", "not a \\1", options = "insensitive")
#' @aliases substitutes s
#' @export re_substitutes substitutes s
re_substitutes <- substitutes <- s <- function(data, pattern, replacement, global = FALSE, options = NULL, ...) {
  pattern <- add_options(pattern, options)
  method <- if (isTRUE(global)) gsub else sub
  method(x = data, pattern = pattern, replacement = replacement, perl = TRUE, ...)
}

add_options <- function(pattern, options) {
  if (!is.null(options)) {
    options <- match_args(options, names(option_map))
    p("(?", p(option_map[options]), ")", pattern)
  }
  else {
    pattern
  }
}

match_args <- function(arg, choices) {
  matches <- pmatch(arg, choices)
  if (any(is.na(matches))) {
    stop(gettextf("'arg' should be one of %s", paste(dQuote(choices),
          collapse = ", ")), domain = NA)
  }
  choices[matches]
}

option_map <- c(
  "insensitive" = "i",
  "multi-line" = "m",
  "single-line" = "s",
  "extended" = "x",
  "ungreedy" = "U"
  )

no_capture <- function(match) {
  is.null(attr(match, "capture.start", exact = TRUE))
}

auto_name <- function(names) {
  missing <- names == ""
  if (all(!missing)) {
    return(names)
  }
  names[missing] <- seq_along(names)[missing]
  names
}

context("re_matches")
df2 <- function(...) {
  data.frame(..., stringsAsFactors = FALSE, check.names = FALSE)
}
string <- c("this is Text", "chr-12", "12343 66544456")

test_that("re_matches if given other than character vector", {

  expect_equal(re_matches(NA, rex(digit)), NA)
  expect_equal(re_matches(1, rex(digit)), TRUE)
  expect_equal(re_matches(1, rex(capture(digit))), df2("1" = "1"))
  expect_equal(re_matches("a", rex(capture(digit))), df2("1" = NA_character_))
})

test_that("re_matches no capture returns a logical", {

  expect_equal(re_matches(string, rex(digit)), c(FALSE, TRUE, TRUE))
  expect_equal(re_matches(string, rex(digits)), c(FALSE, TRUE, TRUE))
  expect_equal(re_matches(string, rex(alpha)), c(TRUE, TRUE, FALSE))
  expect_equal(re_matches(string, rex("-")), c(FALSE, TRUE, FALSE))
})

test_that("re_matches capture returns a data frame", {

  expect_equal(re_matches(string, rex(capture(digits))),
    df2(`1`=c(NA_character_, "12", "12343")))
  expect_equal(re_matches(string, rex(capture(alphas))),
    df2(`1`=c("this", "chr", NA_character_)))

  expect_equal(re_matches(string, rex(capture(alphas), " ", capture(alphas))),
    df2(`1`=c("this", NA_character_, NA_character_),
        `2`=c("is", NA_character_, NA_character_)))
})

test_that("re_matches with global returns a list of matches", {

  # Global without captures doesn"t make a ton of sense but lets test it anyway
  expect_equal(re_matches(string, rex("is"), global = TRUE),
    list(c(TRUE, TRUE), FALSE, FALSE))
})

test_that("re_matches with global returns a list of data.frames", {

  expect_equal(re_matches(string, rex(capture(any_letters, "is")), global = TRUE),
    list(df2(`1` = c("this", "is")), df2(`1` = NA_character_), df2(`1` = NA_character_)) )
  expect_equal(re_matches(string, rex(capture(digits, name = "number")), global = TRUE),
    list(df2("number" = NA_character_), df2("number" = "12"), df2("number" = c("12343", "66544456"))))
})

test_that("re_matches named capture returns named data frame", {

  expect_equal(re_matches(string, rex(capture(digits, name = "numbers"))),
    df2(numbers=c(NA_character_, "12", "12343")))
  expect_equal(re_matches(string, rex(capture(alphas, name = "letters"))),
    df2(letters=c("this", "chr", NA_character_)))
})

test_that("examples are correct", {
string <- c("this is a", "test string")
expect_equal(re_matches(string, rex(capture(alphas, name = "first_word"), space,
                capture(alphas, name = "second_word"))),
            df2(first_word = c("this", "test"),
                second_word = c("is", "string")))

expect_equal(re_matches(string, rex(capture("test"))),
  df2(`1`=c(NA_character_, "test")))
})

context("re_matches - locations")
test_that("re_matches if given other than character vector", {

  expect_equal(re_matches(NA, rex(digit), locations = TRUE),
    df2(start = NA_integer_, end = NA_integer_))

  expect_equal(re_matches(1, rex(digit), locations = TRUE),
    df2(start = 1L, end = 1L))

  expect_equal(re_matches(1, rex(capture(digit)), locations = TRUE),
    df2("1" = "1", "1.start" = 1L, "1.end" = 1L))

  expect_equal(re_matches("a", rex(capture(digit)), locations = TRUE),
    df2("1" = NA_character_, "1.start" = NA_integer_, "1.end" = NA_integer_))
})

test_that("re_matches no capture returns a logical", {

  expect_equal(re_matches(string, rex(digit), locations = TRUE),
    df2(start = c(NA_integer_, 5L, 1L), end = c(NA_integer_, 5L, 1L)))

  expect_equal(re_matches(string, rex(digits), locations = TRUE),
    df2(start = c(NA_integer_, 5L, 1L), end = c(NA_integer_, 6L, 5L)))

  expect_equal(re_matches(string, rex(alpha), locations = TRUE),
    df2(start = c(1L, 1L, NA_integer_), end = c(1L, 1L, NA_integer_)))

  expect_equal(re_matches(string, rex("-"), locations = TRUE),
    df2(start = c(NA_integer_, 4L, NA_integer_), end = c(NA_integer_, 4L, NA_integer_)))
})

test_that("re_matches capture returns a data frame", {

  expect_equal(re_matches(string, rex(capture(digits)), locations = TRUE),
    df2("1" = c(NA_character_, "12", "12343"), "1.start" = c(NA_integer_, 5L, 1L), "1.end" = c(NA_integer_, 6L, 5L)))

  expect_equal(re_matches(string, rex(capture(alphas)), locations = TRUE),
    df2("1" = c("this", "chr", NA_character_), "1.start" = c(1L, 1L, NA_integer_), "1.end" = c(4L, 3L, NA_integer_)))

  expect_equal(re_matches(string, rex(capture(alphas), " ", capture(alphas)), locations = TRUE),
    df2(
      "1" = c("this", NA_character_, NA_character_),
      "1.start" = c(1L, NA_integer_, NA_integer_),
      "1.end" = c(4L, NA_integer_, NA_integer_),

      "2" = c("is", NA_character_, NA_character_),
      "2.start" = c(6L, NA_integer_, NA_integer_),
      "2.end" = c(7L, NA_integer_, NA_integer_)
    ))
})

test_that("re_matches named capture returns named data frame", {

  expect_equal(re_matches(string, rex(capture(digits, name = "numbers")), locations = TRUE),
    df2("numbers" = c(NA_character_, "12", "12343"),
        "numbers.start" = c(NA_integer_, 5L, 1L),
        "numbers.end" = c(NA_integer_, 6L, 5L)))

  expect_equal(re_matches(string, rex(capture(alphas, name = "letters")), locations = TRUE),
    df2("letters" = c("this", "chr", NA_character_),
        "letters.start" = c(1L, 1L, NA_integer_),
        "letters.end" = c(4L, 3L, NA_integer_)))

})

test_that("re_matches with global returns a list of matches", {

  expect_equal(re_matches(string, rex("is"), global = TRUE, locations = TRUE),
    list(
      df2(start = c(3L, 6L), end = c(4L, 7L)),
      df2(start = NA_integer_, end = NA_integer_),
      df2(start = NA_integer_, end = NA_integer_)
      )
    )
})

test_that("re_matches with global returns a list of data.frames", {

  expect_equal(re_matches(string, rex(capture(any_letters, "is")), global = TRUE, locations = TRUE),
    list(
      df2("1" = c("this", "is"), "1.start" = c(1L, 6L), "1.end" = c(4L, 7L)),
      df2("1" = NA_character_, "1.start" = NA_integer_, "1.end" = NA_integer_),
      df2("1" = NA_character_, "1.start" = NA_integer_, "1.end" = NA_integer_)
      )
    )

  expect_equal(re_matches(string, rex(capture(digits, name = "number")), global = TRUE, locations = TRUE),
    list(
      df2(number = NA_character_, number.start = NA_integer_, number.end = NA_integer_),
      df2(number = "12", number.start = 5L, number.end = 6L),
      df2(number = c("12343", "66544456"), number.start = c(1L, 7L), number.end = c(5L, 14L))
      )
    )
})

test_that("examples are correct", {
string <- c("this is a", "test string")
expect_equal(re_matches(string, rex(capture(alphas, name = "first_word"), space,
                capture(alphas, name = "second_word"))),
            df2(first_word = c("this", "test"),
                second_word = c("is", "string")))

expect_equal(re_matches(string, rex(capture("test"))),
  df2(`1`=c(NA_character_, "test")))
})

context("re_substitutes")
test_that("s substitutes properly, with and without options", {
  expect_equal(re_substitutes(string, rex("Text"), "test"),
    c("this is test", "chr-12", "12343 66544456"))

  expect_equal(re_substitutes(string, rex("text"), "test", options="insensitive"),
    c("this is test", "chr-12", "12343 66544456"))

  expect_equal(re_substitutes(string, "i", "x", global = TRUE),
    c("thxs xs Text", "chr-12", "12343 66544456"))
})

context("match_args")
test_that("match args stops if the arg does not match", {
  expect_error(match_args("test", names(option_map)))

  expect_equal(match_args("ungreedy", names(option_map)), "ungreedy")

  expect_equal(match_args("ungre", names(option_map)), "ungreedy")
})

#' Toggles \pkg{rex} mode.
#'
#' While within rex mode, functions used within the \code{\link{rex}} function
#' are attached, so one can get e.g. auto-completion within editors.
#'
#' @export
rex_mode <- function() {

  ## Enter rex mode
  if (!.rex$mode) {
    .rex$mode <- TRUE
    message("Rex functions and shortcuts attached!")
    ## We know what we're doing, so hide the R CMD check note
    suppressMessages(
      eval(call("attach", call("$", as.name(".rex"), as.name("env"))))
    )
  }

  ## Exit rex mode
  else {
    message("Rex functions and shortcuts detached!")
    .rex$mode <- FALSE
    detach(".rex$env")
  }

}

context("rex_mode")
test_that("rex_mode attaches .rex$env to the search_path", {
  expect_false(".rex$env" %in% search())

  rex_mode()

  expect_true(".rex$env" %in% search())

  rex_mode()

  expect_false(".rex$env" %in% search())
})

#' @include utils.R
#' @include escape.R
NULL

#' Or
#'
#' The special binary function \code{\%or\%} can be used to specify a set
#' of optional matches.
#'
#' @rdname or
#' @usage x \%or\% y
#' @param x A string.
#' @param y A string.
#' @family rex
#' @inheritParams capture
`%or%` <- function(x, y) {
  group(p(escape(x)), regex("|"), p(escape(y)))
}
register(`%or%`)

#' describeIn or regular function can also be used, useful for more than 2 arguments.
or <- function(...) {
  group(regex(paste0(collapse = "|", unlist(escape_dots(...)))))
}
register(or)

#' @include escape.R
#' @include utils.R
NULL

#' Lookarounds
#'
#' These functions provide an interface to perl lookarounds.
#'
#' Special binary functions are used to infer an ordering, since often you
#' might wish to match a word / set of characters conditional on the start
#' and end of that word.
#'
#' \itemize{
#'   \item \code{\%if_next_is\%}: \code{TRUE} if x follows y
#'   \item \code{\%if_next_isnt\%}: \code{TRUE} if x does not follow y
#'   \item \code{\%if_prev_is\%}: \code{TRUE} if y comes before x
#'   \item \code{\%if_prev_isnt\%}: \code{TRUE} if y does not come before x
#' }
#' @param x A regex pattern.
#' @param y A regex pattern.
#' @name lookarounds
#' @title Lookarounds
#' @family rex
#' @seealso Perl 5 Documentation \url{http://perldoc.perl.org/perlre.html#Extended-Patterns}
#' @examples
#' stopifnot(grepl(rex("crab" %if_next_is% "apple"), "crabapple", perl = TRUE))
#' stopifnot(grepl(rex("crab" %if_prev_is% "apple"), "applecrab", perl = TRUE))
#' stopifnot(grepl(rex(range("a", "e") %if_next_isnt% range("f", "g")),
#'   "ah", perl = TRUE))
#' stopifnot(grepl(rex(range("a", "e") %if_next_is% range("f", "i")),
#'   "ah", perl = TRUE))
NULL

#' @rdname lookarounds
`%if_next_is%` <- function(x, y) {
  p("(?:", escape(x), "(?=", escape(y), ")", ")")
}
register(`%if_next_is%`)

#' @rdname lookarounds
`%if_next_isnt%` <- function(x, y) {
  p("(?:", escape(x), "(?!", escape(y), ")", ")")
}
register(`%if_next_isnt%`)

#' @rdname lookarounds
`%if_prev_is%` <- function(x, y) {
  p("(?:", "(?<=", escape(y), ")", escape(x), ")")
}
register(`%if_prev_is%`)

#' @rdname lookarounds
`%if_prev_isnt%` <- function(x, y) {
  p("(?:", "(?<!", escape(y), ")", escape(x), ")")
}
register(`%if_prev_isnt%`)

.rex <- new.env(parent = emptyenv())
.rex$env <- new.env(parent = emptyenv())
.rex$mode <- FALSE

register <- function(...) {
  names <- gsub("`", "", as.character(eval(substitute(alist(...)))), fixed = TRUE)

  list2env(structure(list(...), .Names = names), envir = .rex$env)
}

register_object <- function(object) {
  list2env(as.list(object), envir = .rex$env)
}

context("aaa")
test_that("register adds a function to the .rex$env enviornment", {
  a <- identity

  register(a)

  expect_true("a" %in% ls(envir=.rex$env))

  rm("a", envir = .rex$env)
})

test_that("register_object adds all the objects to the .rex$env environment", {
  b <- list(x=1, y = 2)

  register_object(b)

  expect_true("x" %in% ls(envir=.rex$env))

  expect_true("y" %in% ls(envir=.rex$env))

  expect_false("b" %in% ls(envir=.rex$env))

  rm("x", envir = .rex$env)

  rm("y", envir = .rex$env)
})

#' @include utils.R
#' @include escape.R
NULL

#' Create character classes
#'
#' There are multiple ways you can define a character class.

#' @inheritParams capture
#' @param start beginning of character class
#' @param end end of character class
#' @param x text to include in the character class (must be escaped manually)
#' @inheritParams wildcards
#' @export
#' @family rex
#' @examples
#' # grey = gray
#' re <- rex("gr", one_of("a", "e"), "y")
#' grepl(re, c("grey", "gray")) # TRUE TRUE
#'
#' # Match non-vowels
#' re <- rex(none_of("a", "e", "i", "o", "u"))
#' # They can also be in the same string
#' re <- rex(none_of("aeiou"))
#' grepl(re, c("k", "l", "e")) # TRUE TRUE FALSE
#'
#' # Match range
#' re <- rex(range("a", "e"))
#' grepl(re, c("b", "d", "f")) # TRUE TRUE FALSE
#'
#' # Explicit creation
#' re <- rex(character_class("abcd\\["))
#' grepl(re, c("a", "d", "[", "]")) # TRUE TRUE TRUE FALSE
#' @describeIn character_class explicitly define a character class
character_class <- function(x) structure(x, class = c("character_class", "regex"))

#' @describeIn character_class matches one of the specified characters.
one_of <- function(...) {
  p( "[", p(character_class_escape_dots(...)), "]" )
}
register(one_of)

#' @describeIn character_class matches zero or more of the specified characters.
any_of <- function(..., type = c("greedy", "lazy", "possessive")) {
  add_type(p(one_of(...), "*"), type)
}
register(any_of)

#' @describeIn character_class matches one or more of the specified characters.
some_of <- function(..., type = c("greedy", "lazy", "possessive")) {
  add_type(p(one_of(...), "+"), type)
}
register(some_of)

#' @describeIn character_class matches anything but one of the specified characters.
#' @aliases except
none_of <- except <- function(...) {
  p( "[^", p(character_class_escape_dots(...)), "]" )
}
register(none_of, except)

#' @describeIn character_class matches zero or more of anything but the specified characters.
except_any_of <- function(..., type = c("greedy", "lazy", "possessive")) {
  add_type(p(none_of(...), "*"), type)
}
register(except_any_of)

#' @describeIn character_class matches one or more of anything but the specified characters.
except_some_of <- function(..., type = c("greedy", "lazy", "possessive")) {
  add_type(p(none_of(...), "+"), type)
}
register(except_some_of)

#' @describeIn character_class matches one of any of the characters in the range.
range <- function(start, end) {
  character_class(p(character_class_escape(start), "-", character_class_escape(end)))
}
register(range)

#' @describeIn character_class matches one of any of the characters in the range.
`:` <- function(start, end) {
  if (all(is.character(start), is.character(end))) {
    range(start, end)
  } else {
    .Primitive(":")(start, end)
  }
}
register(`:`)

#' @describeIn character_class matches one of any of the characters except those in the range.
exclude_range <- function(start, end) {
  character_class(p("^", character_class_escape(start), "-", character_class_escape(end)))
}
register(exclude_range)

context("one_of")
test_that("simple text is correct", {
  re <- rex(one_of(1:9))

  expect_equal(re, regex("[123456789]"))
  lapply(1:9, function(x) {
    expect_true(grepl(re, x, perl = TRUE), info=x)
  })
  expect_false(grepl(re, "a", perl = TRUE))


  vals <- c("a", "b", "c", 0, 3, 5)

  re <- rex(one_of(vals))

  expect_equal(re, regex("[abc035]"))

  lapply(vals, function(x) {
    expect_true(grepl(re, x, perl = TRUE), info = x)
  })

  expect_false(grepl(re, "d", perl = TRUE))
})

test_that("escapes correctly", {
  vals <- c("[", "]")

  re <- rex(one_of(vals))

  expect_equal(re, regex("[\\[\\]]"))

  lapply(vals, function(x) {
    expect_true(grepl(re, x, perl = TRUE), info = x)
  })

  expect_false(grepl(re, "{", perl = TRUE))
})

tests <- c(quote("a"),
      quote(any),
      quote(quote),
      quote(quotes),
      quote(lower),
      quote(upper),
      quote(list(upper, lower)),
      quote(list("[", "]")))

test_that("any_of equals one_of() plus *", {
  lapply(tests,
    function(x) {
      re1 <- regex(paste0(rex(one_of(eval(x))), "*"))
      re2 <- rex(any_of(eval(x)))

      expect_equal(re1, re2, info = paste(sep=" : ", re1, re2))
    })
})

test_that("some_of equals one_of plus +", {
  lapply(tests,
    function(x) {
      re1 <- regex(paste0(rex(one_of(eval(x))), "+"))
      re2 <- rex(some_of(eval(x)))

      expect_equal(re1, re2, info = paste(sep=" : ", re1, re2))
    })
})

test_that("except_any equals none_of() plus *", {
  lapply(tests,
    function(x) {
      re1 <- regex(paste0(rex(none_of(eval(x))), "*"))
      re2 <- rex(except_any_of(eval(x)))

      expect_equal(re1, re2, info = paste(sep=" : ", re1, re2))
    })
})

test_that("except_some equals none_of() plus +", {
  lapply(tests,
    function(x) {
      re1 <- regex(paste0(rex(none_of(eval(x))), "+"))
      re2 <- rex(except_some_of(eval(x)))

      expect_equal(re1, re2, info = paste(sep=" : ", re1, re2))
    })
})

context("none_of")
test_that("simple text is correct", {
  re <- rex(none_of(1:9))

  expect_equal(re, regex("[^123456789]"))
  lapply(1:9, function(x) {
    expect_false(grepl(re, x, perl = TRUE), info=x)
  })
  expect_true(grepl(re, "a", perl = TRUE))


  vals <- c("a", "b", "c", 0, 3, 5)

  re <- rex(none_of(vals))

  expect_equal(re, regex("[^abc035]"))

  lapply(vals, function(x) {
    expect_false(grepl(re, x, perl = TRUE), info = x)
  })

  expect_true(grepl(re, "d", perl = TRUE))
})

test_that("escapes correctly", {
  vals <- c("[", "]")

  re <- rex(none_of(vals))

  expect_equal(re, regex("[^\\[\\]]"))

  lapply(vals, function(x) {
    expect_false(grepl(re, x, perl = TRUE), info = x)
  })

  expect_true(grepl(re, "{", perl = TRUE))
})

context("range")
test_that("matches basic characters", {
  re <- rex(range(1, 3))

  expect_equal(re, regex("[1-3]"))

  lapply(1:3, function(x) {
    expect_true(grepl(re, x), info=x)
  })

  lapply(4:9, function(x) {
    expect_false(grepl(re, x), info=x)
  })

})
test_that("escapes special characters", {
  re <- rex(range("[", "}"))

  expect_equal(re, regex("[\\[-}]"))

  lapply(c("[", "}"), function(x) {
    expect_true(grepl(re, x), info=x)
  })

})

context("exclude_range")
test_that("matches basic characters", {
  re <- rex(exclude_range(1, 3))

  expect_equal(re, regex("[^1-3]"))

  lapply(1:3, function(x) {
    expect_false(grepl(re, x, perl = TRUE), info=x)
  })

  lapply(4:9, function(x) {
    expect_true(grepl(re, x, perl = TRUE), info=x)
  })

})

context("one_of")
test_that("matches basic characters", {
  expect_equal(rex(one_of("a", "b", "rst")), regex("[abrst]"))
})

test_that("escapes special characters", {
  expect_equal(rex(one_of("^", "b", "\\")), regex("[\\^b\\\\]"))
})

context("except")
test_that("matches basic characters", {
  expect_equal(rex(except("a", "b", "rst")), regex("[^abrst]"))
})

test_that("escapes special characters", {
  expect_equal(rex(except("^", "b")), regex("[^\\^b]"))
})

test_that("none_of is the same as except", {
  expect_equal(rex(none_of("^", "b", 1:10)), rex(except("^", "b", 1:10)))
})

context("character_class")
test_that("examples are correct", {
  # grey = gray
  re <- rex("gr", one_of("a", "e"), "y")
  expect_equal(grepl(re, c("grey", "gray")), c(TRUE, TRUE)) # TRUE TRUE

  # Match non-vowels
  re <- rex(none_of("a", "e", "i", "o", "u"))
  # They can also be in the same string
  re2 <- rex(none_of("aeiou"))
  expect_identical(re, re2)
  expect_equal(grepl(re, c("k", "l", "e")), c(TRUE, TRUE, FALSE)) # TRUE TRUE FALSE

  # Match range
  re <- rex(range("a", "e"))
  expect_equal(grepl(re, c("b", "d", "f")), c(TRUE, TRUE, FALSE)) # TRUE TRUE FALSE

  # Explicit creation (note you have to escape manually here)
  re <- rex(character_class("abcd\\["))
  expect_equal(grepl(re, c("a", "d", "[", "]")), c(TRUE, TRUE, TRUE, FALSE)) # TRUE TRUE TRUE FALSE
})
test_that("escapes special characters", {
  re <- rex(exclude_range("[", "}"))

  expect_equal(re, regex("[^\\[-}]"))

  lapply(c("[", "}"), function(x) {
    expect_false(grepl(re, x, perl = TRUE), info=x)
  })

  expect_true(grepl(re, "A", perl = TRUE))
})

#' @include escape.R
#' @include character_class.R
#' @include utils.R
NULL

#' Generate a regular expression.
#' @export
#' @family rex
#' @param ... \code{\link{shortcuts}}, R variables, text, or other \pkg{rex}
#' functions.
#' @param env environment to evaluate the rex expression in.
#' @aliases rex_
rex <- function(..., env = parent.frame()) {
  args <- lazyeval::lazy_dots(...)
  rex_(args, env)
}

#' @export
rex_ <- function(args, env = parent.frame()) {

  args <- lazyeval::as.lazy_dots(args, env)

  # this needs the as.list because eval only looks at the enclos if envir is
  # not an environment
  regex(p(escape(lazyeval::lazy_eval(args, as.list(.rex$env)))))
}

#' @describeIn regex coerce regex object to a character
#' @export
as.character.regex <- function(x, ...) x

#' Coerce objects to a \code{\link{regex}}.
#' @name as.regex
#' @param x Object to coerce to \code{\link{regex}}.
#' @param ... further arguments passed to methods.
#' @export
as.regex <- function(x, ...) UseMethod("as.regex")

#' @export
#' @describeIn as.regex Simply escape the Object.
as.regex.default <- function(x, ...) escape(x)

#' @export
#' @describeIn regex Print regex object
print.regex <- function(x, ...){
  cat(paste(strwrap(x), collapse = "\n"), "\n", sep = "")
}

#' Regular Expression
#'
#' Specify an explicit regular expression.  This expression must already be
#' escaped.
#' @param x Object
#' @param ... further arguments
#' @seealso \code{\link{as.regex}} to coerce to a regex object.
#' @export
regex <- function(x, ...) structure(x, class = "regex")

#' Register the Rex shortcuts
#'
#' If you are using rex in another package you need to call this function to
#' register all of the rex shortcuts so that spurious NOTEs about global
#' variables being generated during R CMD check.
#' @param pkg_name the package to register the shortcuts in
#' @export
register_shortcuts <- function(pkg_name) {
  invisible(utils::globalVariables(ls(.rex$env), pkg_name))
}

#' @export
c.regex <- function(..., recursive = FALSE) {
  regex(c(unlist(lapply(list(...), unclass))))
}

context("or")
test_that("%or% creates an alternation", {
  re <- rex("w", "x" %or% "y", "z")
  expect_equal(re, regex("w(?:x|y)z"))

  expect_true(grepl(re, "wxz"))
  expect_true(grepl(re, "wyz"))
  expect_false(grepl(re, "waz"))
})

test_that("or with multiple inputs works", {
  re <- rex(or("x", "yx", "z"))
  expect_equal(re, regex("(?:x|yx|z)"))
  lapply(c("x", "yx", "z"), function(x) {
    expect_true(grepl(re, x, perl=TRUE), info=x)
  })

  expect_false(grepl(re, c("y")))
  expect_false(grepl(re, c("a")))
})
test_that("or with variable inputs works", {
  variable <- c("test", "variable")
  re <- rex(or(variable))

  expect_equal(re, regex("(?:test|variable)"))

  lapply(variable, function(x){
    expect_true(grepl(re, x), info=x)
  })
})

context("capture")
test_that("matches basic characters", {
  x <- "text"
  re <- rex(capture(x))

  expect_equal(re, regex("(text)"))

  expect_true(grepl(re, x))

  expect_equal(gsub(re, "\\1", x), x)

  expect_equal(gsub(re, "replacement", x), "replacement")
})

test_that("escapes special characters", {
  x <- "^[x$\\"
  re <- rex(capture(x))

  expect_equal(re, regex("(\\^\\[x\\$\\\\)"))

  expect_true(grepl(re, x))

  expect_equal(gsub(re, "\\1", x), x)

  expect_equal(gsub(re, "replacement", x), "replacement")
})

test_that("examples work", {
 re <- rex(
   # first quotation mark
   capture(quote),

   # match all non-matching quotation marks
   zero_or_more(except(capture_group(1))),

   # end quotation mark (matches first)
   capture_group(1)
 )

 expect_equal(re, regex("(['\"])(?:[^\\g{1}])*\\g{1}"))

 lapply(c("\"\"", "\"'\"", "\"arst\"", "''", "'arst'", "'\"'"),
   function(x) {
     expect_true(grepl(re, x, perl = TRUE), info=x)
   }
 )

 lapply(c("'a", "'asr\""), function(x) {
    expect_false(grepl(re, x, perl = TRUE), info = x)
 })

})

context("named capture")
test_that("examples work", {
  re <- rex(
    capture(name = "fruit", or("apple", "orange")),
    "=",
    capture_group("fruit")
    )

  expect_true(grepl(re, "apple=apple", perl=TRUE))
  expect_true(grepl(re, "orange=orange", perl=TRUE))
  expect_false(grepl(re, "apple=orange", perl=TRUE))
})

context("lookarounds")
test_that("lookarounds work", {

  r <- rex(
    start, "foo" %if_next_isnt% "bar"
  )

  expect_true(grepl(r, "fooba", perl = TRUE))
  expect_false(grepl(r, "foobar", perl = TRUE))
})
context("if_next_is")
test_that("matches basic characters", {
  re <- rex("a" %if_next_is% "b")
  expect_equal(re, regex("(?:a(?=b))"))
  expect_true(grepl(re, "ab", perl=TRUE))
  expect_false(grepl(re, "ac", perl=TRUE))
})
test_that("escapes special characters", {
  re <- rex("[" %if_next_is% "?=")
  expect_equal(re, regex("(?:\\[(?=\\?=))"))
  expect_true(grepl(re, "[?=", perl=TRUE))
  expect_false(grepl(re, "?=[", perl=TRUE))
})

context("if_next_isnt")
test_that("matches basic characters", {
  re <- rex("a" %if_next_isnt% "b")
  expect_equal(re, regex("(?:a(?!b))"))
  expect_true(grepl(re, "ac", perl=TRUE))
  expect_false(grepl(re, "ab", perl=TRUE))
})
test_that("escapes special characters", {
  re <- rex("[" %if_next_isnt% "?=")
  expect_equal(re, regex("(?:\\[(?!\\?=))"))
  expect_true(grepl(re, "?=[", perl=TRUE))
  expect_false(grepl(re, "[?=", perl=TRUE))
})

context("if_prev_is")
test_that("matches basic characters", {
  re <- rex("a" %if_prev_is% "b")
  expect_equal(re, regex("(?:(?<=b)a)"))
  expect_true(grepl(re, "ba", perl=TRUE))
  expect_false(grepl(re, "ab", perl=TRUE))
})
test_that("escapes special characters", {
  re <- rex("[" %if_prev_is% "<=?")
  expect_equal(re, regex("(?:(?<=<=\\?)\\[)"))
  expect_true(grepl(re, "<=?[", perl=TRUE))
  expect_false(grepl(re, "[b", perl=TRUE))
})

context("if_prev_isnt")
test_that("matches basic characters", {
  re <- rex("a" %if_prev_isnt% "b")
  expect_equal(re, regex("(?:(?<!b)a)"))
  expect_true(grepl(re, "ab", perl=TRUE))
  expect_false(grepl(re, "ba", perl=TRUE))
})
test_that("escapes special characters", {
  re <- rex("[" %if_prev_isnt% "!<?")
  expect_equal(re, regex("(?:(?<!!<\\?)\\[)"))
  expect_true(grepl(re, "[b", perl=TRUE))
  expect_false(grepl(re, "!<?[", perl=TRUE))
})

#' @include escape.R
#' @include utils.R
NULL

#' Wildcards
#'
#' @inheritParams capture
#' @param type the type of match to perform.
#'
#' There are three match types
#' \enumerate{
#'   \item \code{greedy}: match the longest string.  This is the default matching type.
#'   \item \code{lazy}: match the shortest string.  This matches the shortest string from the same anchor point, not necessarily the shortest global string.
#'   \item \code{possessive}: match and don't allow backtracking
#' }
#' @family rex
#' @name wildcards
NULL

#' @describeIn wildcards match \code{...} zero or more times.
zero_or_more <- function(..., type = c("greedy", "lazy", "possessive")) {
  add_type(p("(?:", p(escape_dots(...)), ")*"), type)
}
register(zero_or_more)

#' @describeIn wildcards match \code{...} one or more times.
one_or_more <- function(..., type = c("greedy", "lazy", "possessive")) {
  add_type(p("(?:", p(escape_dots(...)), ")+"), type)
}
register(one_or_more)

#' @describeIn wildcards match \code{...} zero or one times.
#' @aliases zero_or_one
maybe <- zero_or_one <- function(..., type = c("greedy", "lazy", "possessive")) {
  p("(?:", p(escape_dots(...)), ")?")
}
register(maybe, zero_or_one)

add_type <- function(x, type = c("greedy", "lazy", "possessive")) {
  type <- match.arg(type)

  switch(type,
    greedy = x,
    lazy = p(x, "?"),
    possessive = p(x, "+")
  )
}

#' Do not match
#'
#' @inheritParams capture
#' @inheritParams zero_or_more
#' @family rex
# This is slightly different than if_next_isn't because we want to match
# anything that is not the search term as well
not <- function(..., type = c("greedy", "lazy", "possessive")) {
  add_type(p("(?:(?!", escape_dots(...), ").)*"), type = type)
}
register(not)

context("rex")

test_that("start works", {

  r <- rex(
    start, letter
    )

  expect_true(grepl(r, "abcdef"))
  expect_false(grepl(r, "123456", perl = TRUE))

})

test_that("end works", {

  r <- rex(
    "Z", end
    )

  expect_true(grepl(r, "abcZ"))
  expect_false(grepl(r, "abc"))

})

test_that("version parsing works", {

  r <- rex(
    start,
    capture(numbers),
    any,
    capture(numbers),
    any,
    capture(numbers),
    any,
    capture(numbers),
    end
    )

  expect_identical(
    gsub(r, "\\1 \\2 \\3 \\4", "3.1.1-1", perl = TRUE),
    "3 1 1 1"
    )

})

test_that("verbs in rex work", {

  r <- rex(
    start, "foo", zero_or_more(any), "bar", end
    )

  expect_true(grepl(r, "fooABCbar", perl = TRUE))
  expect_true(grepl(r, "foo123\tbar", perl = TRUE))

})

test_that("Simple URL parsing works", {

  # TODO: get these working better again
  ## Decompose a URL into its components.
  ## Example by LT (http://www.cs.uiowa.edu/~luke/R/regexp.html).
  x <- "http://stat.umn.edu:80/xyz"
  re <- "^(?:(((?:(?:(?!:).)*)+)://))?((?:(?:(?!:/).)*)+)(?:(:((?:[[:digit:]]+)+)))?(?:(/(?:.)*))?$"
  #m <- regexec(re, x)
  #m
  #regmatches(x, m)
  ## Element 3 is the protocol, 4 is the host, 6 is the port, and 7
  ## is the path.  We can use this to make a function for extracting the
  ## parts of a URL:
  URL_parts <- function(x) {
    m <- regexec(re, x)
    parts <- do.call(rbind,
      lapply(regmatches(x, m), `[`, c(3L, 4L, 6L, 7L)))
    colnames(parts) <- c("protocol","host","port","path")
    parts
  }
  #URL_parts(x)

  r <- rex(

    start,

    ## match the protocol -- may exist or may not
    maybe(capture(
        capture(except_some_of(":")),
        "://"
        )),

    ## match the path
    capture(one_or_more(not(":/"))),

    ## get the port
    maybe(capture(":", capture(numbers))),

    ## and the rest
    maybe(capture("/", anything)),

    end

    )

  rbind(r = r, m = re)
  n <- gregexpr(r, x, perl = TRUE)[[1]]
  split_matches <- function(string, matches) {
    starts <- attr(matches, "capture.start")
    lengths <- attr(matches, "capture.length")
    ends <- starts + lengths - 1
    c(string, substring(string, starts, ends))
  }

  #split_matches(x, n)
  #regmatches(x, m)[[1]]
  #expect_equal(regmatches(x, m)[[1]], split_matches(x, n))

})

context("URL Validation")
test_that("URL Validation works", {
  valid_chars <- rex(except_some_of(".", "/", " ", "-"))

  `%>%` <- magrittr::`%>%`

  re <- rex(
    start,

    # protocol identifier (optional) + //
    group(list("http", maybe("s")) %or% "ftp", "://"),

    # user:pass authentication (optional)
    maybe(non_spaces,
      maybe(":", zero_or_more(non_space)),
      "@"),

    #host name
    group(zero_or_more(valid_chars, zero_or_more("-")), one_or_more(valid_chars)),

    #domain name
    zero_or_more(".", zero_or_more(valid_chars, zero_or_more("-")), one_or_more(valid_chars)),

    #TLD identifier
    group(".", valid_chars %>% at_least(2)),

    # server port number (optional)
    maybe(":", digit %>% between(2, 5)),

    # resource path (optional)
    maybe("/", non_space %>% zero_or_more()),

    end
    )

  good <- c("http://foo.com/blah_blah",
    "http://foo.com/blah_blah/",
    "http://foo.com/blah_blah_(wikipedia)",
    "http://foo.com/blah_blah_(wikipedia)_(again)",
    "http://www.example.com/wpstyle/?p=364",
    "https://www.example.com/foo/?bar=baz&inga=42&quux",
    "http://✪df.ws/123",
    "http://userid:password@example.com:8080",
    "http://userid:password@example.com:8080/",
    "http://userid@example.com",
    "http://userid@example.com/",
    "http://userid@example.com:8080",
    "http://userid@example.com:8080/",
    "http://userid:password@example.com",
    "http://userid:password@example.com/",
    "http://➡.ws/䨹",
    "http://⌘.ws",
    "http://⌘.ws/",
    "http://foo.com/blah_(wikipedia)#cite-1",
    "http://foo.com/blah_(wikipedia)_blah#cite-1",
    "http://foo.com/unicode_(✪)_in_parens",
    "http://foo.com/(something)?after=parens",
    "http://☺.damowmow.com/",
    "http://code.google.com/events/#&product=browser",
    "http://j.mp",
    "ftp://foo.bar/baz",
    "http://foo.bar/?q=Test%20URL-encoded%20stuff",
    "http://مثال.إختبار",
    "http://例子.测试",
    "http://-.~_!$&'()*+,;=:%40:80%2f::::::@example.com",
    "http://1337.net",
    "http://a.b-c.de",
    "http://223.255.255.254")

  bad <- c(
    "http://",
    "http://.",
    "http://..",
    "http://../",
    "http://?",
    "http://??",
    "http://??/",
    "http://#",
    "http://##",
    "http://##/",
    "http://foo.bar?q=Spaces should be encoded",
    "//",
    "//a",
    "///a",
    "///",
    "http:///a",
    "foo.com",
    "rdar://1234",
    "h://test", # nolint
    "http:// shouldfail.com",
    ":// should fail",
    "http://foo.bar/foo(bar)baz quux",
    "ftps://foo.bar/",
    "http://-error-.invalid/",
    "http://-a.b.co",
    "http://a.b-.co",
    "http://0.0.0.0",
    "http://3628126748",
    "http://.www.foo.bar/",
    "http://www.foo.bar./",
    "http://.www.foo.bar./")

  lapply(good, function(x) {
    expect_true(grepl(re, x, perl = TRUE), info=x)
    })
  lapply(bad, function(x) {
    expect_false(grepl(re, x, perl = TRUE), info=x)
    })
})

context("start")
test_that("matches basic characters", {
  expect_equal(rex(start, "f"), regex("^f"))
})

test_that("escapes special characters", {
  expect_equal(rex(start, "."), regex("^\\."))
})

test_that("matches basic characters", {
  expect_equal(rex(start, "x" %>% n_times(3)), regex("^(?:x){3}"))
})

test_that("matches special identifiers", {
  expect_equal(rex(start, number %>% n_times(2)), regex("^(?:[[:digit:]]){2}"))
})

context("appending expressions")
test_that("adds basic characters", {
  expect_equal(rex("x", "y", "z"),
    regex("xyz"))
  expect_equal(rex("x", maybe("y"), "z"),
    regex("x(?:y)?z"))
})

test_that("escapes special characters", {
  expect_equal(rex(numbers %>% between(0, 2), ".", "$", end),
    regex("(?:[[:digit:]]+){0,2}\\.\\$$"))
})

context("shortcuts - end")
test_that("matches basic characters", {
  expect_equal(rex("x", "y", end),
    regex("xy$"))
})


test_that("escapes special characters", {
  expect_equal(rex("x", "$", end),
    regex("x\\$$"))
})

context("general regex")
test_that("returns a well-formed regex", {
  expect_equal(rex(start, "w", "x" %or% "y", "z", end),
    regex("^w(?:x|y)z$"))
})

context("rex examples")
re <-
  rex(start,
    number %>% n_times(3),
    "-",
    letter %>% n_times(2),
    maybe("#"),
    "a" %or% "b",
    "c" %>% between(2, 4),
    "$",
    end
    )

expect_true(grepl(re, "123-xy#accc$", perl=TRUE))
expect_true(grepl(re, "999-dfbcc$"))
expect_false(grepl(re, "000-df#baccccccccc$"))
expect_false(grepl(re, "444-dd3ac$"))

context("issues")
test_that("#11 Modifiers and named character classes", {
  p <- rex(none_of(alpha))
  expect_true(grepl(p, "6", perl = TRUE))
  })

context("zero_or_more")
test_that("recognizes basic characters", {
  expect_equal(rex(zero_or_more("a"), "b"),
    regex("(?:a)*b"))
})

test_that("recognizes special identifiers", {
  expect_equal(rex(zero_or_more(number), "b"),
    regex("(?:[[:digit:]])*b"))
})

test_that("recognizes shortcuts", {
  re <- rex(zero_or_more(number, type = "lazy"), "E")
  expect_equal(re,
    regex("(?:[[:digit:]])*?E"))

  expect_equal(regmatches(m=regexpr(re, "123EEE", perl = TRUE), "123EEE"), "123E")
})

context("one_or_more")
test_that("recognizes basic characters", {
  expect_equal(rex(one_or_more("a"), "b"),
    regex("(?:a)+b"))
})

test_that("recognizes special identifiers", {
  expect_equal(rex(one_or_more(letter), "b"),
    regex("(?:[[:alpha:]])+b"))
})


context("maybe")
test_that("recognizes basic characters", {
    re <- rex("x", maybe("y"), "z")
    expect_equal(re, regex("x(?:y)?z"))
    expect_true(grepl(re, "xyz"))
    expect_true(grepl(re, "xz"))
})

context("not")
test_that("creates a negative lookahead", {
  re <- rex("x", not("y"), "z")
  expect_equal(re, regex("x(?:(?!y).)*z"))
  expect_true(grepl(re, "xazbc", perl=TRUE))
  expect_true(grepl(re, "xxzabc", perl=TRUE))
  expect_false(grepl(re, "xyzabc", perl=TRUE))

  re <- rex("x432", not("yars"), "tsrz")
  expect_true(grepl(re, "x432tsrz", perl=TRUE))
  expect_true(grepl(re, "x432yartsrz", perl=TRUE))
  expect_false(grepl(re, "x432yarstsrz", perl=TRUE))

  re <- rex(start, not("ars"), "x432")
  expect_true(grepl(re, "x432", perl=TRUE))
  expect_true(grepl(re, "arx432", perl=TRUE))
  expect_false(grepl(re, "arsx432", perl=TRUE))
})

#' ---
#' title: "Stackoverflow Usage Examples"
#' author: "Jim Hester"
#' date: "`r Sys.Date()`"
#' output: rmarkdown::html_vignette
#' vignette: >
#'   %\VignetteIndexEntry{Stackoverflow Usage Examples}
#'   %\VignetteEngine{knitr::rmarkdown}
#'   \usepackage[utf8]{inputenc}
#' ---

#+ cache=FALSE, include=FALSE
knitr::opts_chunk$set(collapse = TRUE, comment = "#>")
knitr::render_markdown(strict = TRUE)

#' ### [http://stackoverflow.com/questions/27106552][]

#' Using [rex](http://cran.r-project.org/web/packages/rex/index.html) may make this type of task a little simpler.

x <- c("a=1", "b=3", "a=9", "c=2", "b=4", "a=2")

#' First extract the names and values from the strings.

library(rex)
matches <- re_matches(x,
  rex(
    capture(name="name", letter),
    "=",
    capture(name="value", digit)
    ))
matches

#' Then tally the groups using `split()`.

groups <- split(as.numeric(matches$value), matches$name)
groups

#' If we try to convert directly to a data.frame from `split()` the groups with
#' fewer members will have their members recycled rather than `NA`, so instead
#' explicitly fill with `NA`.

largest_group <- max(sapply(groups, length))
largest_group

groups <- lapply(groups, function(group) {
  if (length(group) < largest_group) {
    group[largest_group] <- NA
  }
  group
})
groups

#' Finally we can create the data.frame

do.call("data.frame", groups)

#' ### [http://stackoverflow.com/questions/14146362/][] ###

#' Using [rex](http://cran.r-project.org/web/packages/rex/index.html) may make this type of task a little simpler.

mystrings <- c("X2/D2/F4",
               "X10/D9/F4",
               "X3/D22/F4",
               "X9/D22/F9")

library(rex)
matches <- re_matches(mystrings,
  rex(
    "/",
    any,
    capture(name = "numbers", digits)
    )
  )
as.numeric(matches$numbers)

#' ### [http://stackoverflow.com/questions/8613237/][] ##
#' Using [rex](http://cran.r-project.org/web/packages/rex/index.html) may make this type of task a little simpler.

j <- "What kind of cheese isn't your cheese? (wonder) Nacho cheese! (groan) (Laugh)"

library(rex)
matches <- re_matches(j,
  rex(
    "(",
    capture(name = "text", except_any_of(")")),
    ")"),
  global = TRUE)
matches

#' ### [http://stackoverflow.com/questions/22976472][]
#' Using [rex](http://cran.r-project.org/web/packages/rex/index.html) may make this type of task a little simpler.

txt <- as.character("this is just a test! i'm not sure if this is O.K. or if it will work? who knows. regex is sorta new to me..  There are certain cases that I may not figure out??  sad!  ^_^")

re <- rex(
  capture(name = "first_letter", alnum),
  capture(name = "sentence",
    any_non_puncts,
    zero_or_more(
      group(
        punct %if_next_isnt% space,
        any_non_puncts
        )
      ),
    maybe(punct)
    )
  )

re_substitutes(txt, re, "\\U\\1\\E\\2", global = TRUE)

#' ### [http://stackoverflow.com/questions/27172007][]
#' Using [rex](http://cran.r-project.org/web/packages/rex/index.html) may make this type of task a little simpler.

x <- data.frame(
  locationid = c(
    1073744023,
    1073744022,
    1073744025,
    1073744024,
    1073744021,
    1073744026
    ),
  address = c(
    "525 East 68th Street, New York, NY      10065, USA",
    "270 Park Avenue, New York, NY 10017, USA",
    "Rockefeller Center, 50 Rockefeller Plaza, New York, NY 10020, USA",
    "1251 Avenue of the Americas, New York, NY 10020, USA",
    "1301 Avenue of the Americas, New York, NY 10019, USA",
    "44 West 45th Street, New York, NY 10036, USA"
    ))

library(rex)

sep <- rex(",", spaces)

re <-
  rex(
    capture(name = "address",
      except_some_of(",")
    ),
    sep,
    capture(name = "city",
      except_some_of(",")
    ),
    sep,
    capture(name = "state",
      uppers
    ),
    spaces,
    capture(name = "zip",
      some_of(digit, "-")
    ),
    sep,
    capture(name = "country",
      something
    ))

re_matches(x$address, re)

#' ### [http://stackoverflow.com/questions/27155297/][]
#' Using [rex](http://cran.r-project.org/web/packages/rex/index.html) may make this type of task a little simpler.

library(rex)
x <- c(
"https://support.google.com/blogger/topic/12457
https://support.google.com/blogger/topic/12457.
https://support.google.com/blogger/topic/12457] 
<<https://support.google.com/blogger/topic/12457>>
https://support.google.com/blogger/topic/12457,
https://support.google.com/blogger/topic/12457),
xxxxxxhttps://support.google.com/blogger/topic/12457),hhhththta")

re <- rex(
  capture(name = "url",
    "https://support.google.com/blogger/topic/",
    digits
    ))

re_matches(x, re, global = TRUE)[[1]]

#' ### [http://stackoverflow.com/questions/27219421][]
#' Using [rex](http://cran.r-project.org/web/packages/rex/index.html) may make this type of task a little simpler.
tmp <- c("Little Street","A323", "Essex Road (A43)", "M43","Orange street","M4","B2045","New Street")

library(rex)
classify_road <- function(x) {
  res <- re_matches(x,
    rex(
      capture(name = "type",
        upper
      ),
      digit
    )
  )

  res$type[ is.na(res$type) ] <- "Minor"
  paste(res$type, "Road")
}

classify_road(tmp)

#' ### [http://stackoverflow.com/questions/22247410][]
#' Using [rex](http://cran.r-project.org/web/packages/rex/index.html) may make this type of task a little simpler.

x <- "this is a multiline text 
          some more test here 
          before we get to the good stuff 
          \\end{figure}"

re <- rex("\\end{figure}")
re_matches(x, re)

regexpr(re, x, perl = TRUE)

#' ### [http://stackoverflow.com/questions/23447261][]
#' Using [rex](http://cran.r-project.org/web/packages/rex/index.html) may make this type of task a little simpler.

x = structure(list(text = structure(c(4L, 6L, 1L, 2L, 5L, 3L), .Label =     c("ãããæããããéãããæãããInappropriate announce:-(", 
"@AirAsia your direct debit (Maybank) payment gateways is not working. Is it something     you are working to fix?", 
"@AirAsia Apart from the slight delay and shortage of food on our way back from Phuket, both flights were very smooth. Kudos :)", 
"RT @AirAsia: ØØÙØÙÙÙÙ ÙØØØ ØØØÙ ÙØØØØÙ ØØØØÙÙÙí í Now you can enjoy a #great :D breakfast onboard with our new breakfast meals! :D", 
"xdek ke flight @AirAsia Malaysia to LA... hahah..:p bagi la promo murah2 sikit, kompom aku beli...", 
"You know there is a problem when customer service asks you to wait for 103 minutes and your no is 42 in the queue. X-("
), class = "factor"), created = structure(c(5L, 4L, 4L, 3L, 2L, 
1L), .Label = c("1/2/2014 16:14", "1/2/2014 17:00", "3/2/2014 0:54", 
"3/2/2014 0:58", "3/2/2014 1:28"), class = "factor")), .Names = c("text", 
"created"), class = "data.frame", row.names = c(NA, -6L))

emots <- as.character(outer(c(":", ";", ":-", ";-"), c(")", "(", "]", "[", "D", "o", "O", "P", "p"), paste0))

re_matches(x$text,
  rex(
    capture(name = "emoticons",
      or(emots)
    )
  ),
  global = T)

#' ### [http://stackoverflow.com/questions/27234040][]
#' Using [rex](http://cran.r-project.org/web/packages/rex/index.html) may make this type of task a little simpler.

z <- "<TABLE ALIGN=\"RIGHT\" BORDER CELLSPACING=\"0\" CELLPADDING=\"0\">
   <CAPTION><B>MESA HIGH VICTORIES</B></CAPTION>
   <TH>Team</TH>
   <TH>Score</TH>
   <TR ALIGN=\"CENTER\">
   <TD><B>Parkfield High Demons</B></TD>
   <TD><B>28 to 21</B></TD>
   </TR>
   <TR ALIGN=\"CENTER\">
   <TD><B>Burns High Badgers</B></TD>
   <TD><B>14 to 13</B></TD>
   </TR>
   </TABLE>"

re_matches(z,
  rex(
    capture(name="table",
      "<TABLE", zero_or_more(any, type = "lazy"), "<TR"
    )
  ), options="single-line")


#' ### [http://stackoverflow.com/questions/27236435][]
#' Using [rex](http://cran.r-project.org/web/packages/rex/index.html) may make this type of task a little simpler.
x <- "John a11|a12|\n  Ana a21|a22|\n  Jake a31|a23|\n   "

re_matches(x,
  rex(
      any_spaces,
      capture(name = "text",
        except_some_of("|")
      ),
      any_spaces),
  global = TRUE)[[1]]

#' ### [http://stackoverflow.com/questions/25764839][]
#' Using [rex](http://cran.r-project.org/web/packages/rex/index.html) may make this type of task a little simpler.
x <- "MSGSRRKATPASRTRVGNYEMGRTLGEGSFAKVKYAKNTVTGDQAAIKILDREKVFRHKMVEQLKREISTMKLIKHPNVVEIIEVMASKTKIYIVLELVNGGELFDKIAQQGRLKEDEARRYFQQLINAVDYCHSRGVYHRDLKPENLILDANGVLKVSDFGLSAFSRQVREDGLLHTACGTPNYVAPEVLSDKGYDGAAADVWSCGVILFVLMAGYLPFDEPNLMTLYKRICKAEFSCPPWFSQGAKRVIKRILEPNPITRISIAELLEDEWFKKGYKPPSFDQDDEDITIDDVDAAFSNSKECLVTEKKEKPVSMNAFELISSSSEFSLENLFEKQAQLVKKETRFTSQRSASEIMSKMEETAKPLGFNVRKDNYKIKMKGDKSGRKGQLSVATEVFEVAPSLHVVELRKTGGDTLEFHKVCDSFYKNFSSGLKDVVWNTDAAAEEQKQ"
re_matches(x,
  rex(
    capture(name = "amino_acids",
      n(any, 6),
      "K",
      n(any, 6)
      )
    ),
  global = TRUE)[[1]]

locs <- re_matches(x,
  rex(
    "K" %if_prev_is% n(any, 6) %if_next_is% n(any, 6)
    ),
  global = TRUE, locations = TRUE)[[1]]

substring(x, locs$start - 6, locs$end + 6)

#' ### [http://stackoverflow.com/questions/15954171][]
#' Using [rex](http://cran.r-project.org/web/packages/rex/index.html) may make this type of task a little simpler.
x <- c("System configuration: lcpu=96 mem=196608MB ent=16.00")

library(rex)
val <- as.numeric(
  re_matches(x,
    rex("ent=",
      capture(name = "ent", some_of(digit, "."))
      )
    )$ent
  )

#' ### [http://stackoverflow.com/questions/27273996][]
#' Using [rex](http://cran.r-project.org/web/packages/rex/index.html) to construct the regular expression may make it more understandable.
x <- c("_A00_A1234B_", "_A00_A12345B_", "_A1_A12345_")

#' approach #1, assumes always is between the second underscores.
re_matches(x,
  rex(
    "_",
    anything,
    "_",
    capture(anything),
    "_"
  )
)

#' approach #2, assumes an alpha, followed by 4 or 5 digits with a possible trailing alpha.
re_matches(x,
  rex(
    capture(
      alpha,
      between(digit, 4, 5),
      maybe(alpha)
    )
  )
)

#' ### [http://stackoverflow.com/questions/27238323][]
#' Using [rex](http://cran.r-project.org/web/packages/rex/index.html) may make this type of task a little simpler.
string = c("ABC3JFD456", "ARST4DS324")

re_matches(string,
  rex(
    capture(name = "first_number", digit)
    )
  )

#' ### <http://stackoverflow.com/questions/27252250> ###
#' Using [rex](http://cran.r-project.org/web/packages/rex/index.html) may make this type of task a little simpler.
df <- structure(list(Object = c("T00055", "T00055", "E00336", "E00336",
"E00336", "E00336", "T 00054"), Coding = c("T 00055_005_<002_+",
"T 00055_008_<002_+", "E 00336_041_<001_+001_+", "E 00336_041_<001_+001_+001_+",
"E 00336_041_<001_+001_+002_+", "E 00336_041_<001_+001_+002_<",
"T 00054_013_<003_<015_+003_<001_<"), Fn = c(2L, 2L, 3L, 4L,
4L, 4L, 4L), Remaining = c(30L, 30L, 0L, 10L, 56L, 52L, 52L)), .Names = c("Object",
"Coding", "Fn", "Remaining"), row.names = c(NA, -7L), class = "data.frame")

subset(df, grepl(rex(at_least(group("_+", anything), 2)), Coding))

#' ### <http://stackoverflow.com/questions/27195734>
#' Using [rex](http://cran.r-project.org/web/packages/rex/index.html) may make this type of task a little simpler.

ids <- c("367025001", "CT_341796001", "M13X01692-01", "13C025050901", "13C00699551")

re_substitutes(ids,
  rex(non_digits %or% list("01", end)),
  "",
  global = TRUE)

#' ### <http://stackoverflow.com/questions/27237233>
#' Using [rex](http://cran.r-project.org/web/packages/rex/index.html) may make this type of task a little simpler.
library("rvest")
library("stringr")

minimal <- html("<!doctype html><title>blah</title> <p>&nbsp;foo")

bodytext <- minimal %>%
  html_node("body") %>%
  html_text

re_substitutes(bodytext, rex(spaces), "", global = TRUE)

#' ### <http://stackoverflow.com/questions/27227229>
#+ message=FALSE
string <- "this\\(system) {is} [full]."
library(Hmisc)
gsub("\\\\(.)", "\\1", escapeRegex(string))

#' Alternatively [rex](http://cran.r-project.org/web/packages/rex/index.html) may make this type of task a little simpler.
library(rex)
re_substitutes(escape(string), rex("\\", capture(any)), "\\1", global = TRUE)

#' ### <http://stackoverflow.com/questions/27317497>
#' [rex](http://cran.r-project.org/web/packages/rex/) has a [vignette for parsing server logs](http://cran.r-project.org/web/packages/rex/vignettes/log_parsing.html). While the format is not exactly the same as your log you should be able to adapt it to your case fairly easily.
#' As far as reading the log in assuming the file fits in memory your best bet is to read the whole file first with `readLines()`, then the following will put each field into a `data.frame` column.
x <- "Feb  6 12:14:14 localhost haproxy[14389]: 10.0.1.2:33317 [06/Feb/2009:12:14:14.655] http-in static/srv1 10/0/30/69/109 200 2750 - - ---- 1/1/1/1/0 0/0 {1wt.eu} {} \"GET /index.html HTTP/1.1\""
library(rex)
re <- rex(

  capture(name = "process_name", alpha),
  "[",
    capture(name = "pid", digits),
  "]:",
  spaces,
  capture(name = "client_ip", any_of(digit, ".")),
  ":",
  capture(name = "client_port", digits),
  spaces,
  "[",
    capture(name = "accept_date", except_some_of("]")),
  "]",
  spaces,
  capture(name = "frontend_name", non_spaces),
  spaces,
  capture(name = "backend_name", except_some_of("/")),
  "/",
  capture(name = "server_name", non_spaces),
  spaces,
  capture(name = "Tq", some_of("-", digit)),
  "/",
  capture(name = "Tw", some_of("-", digit)),
  "/",
  capture(name = "Tc", some_of("-", digit)),
  "/",
  capture(name = "Tr", some_of("-", digit)),
  "/",
  capture(name = "Tt", some_of("+", digit)),
  spaces,
  capture(name = "status_code", digits),
  spaces,
  capture(name = "bytes_read", some_of("+", digit)),
  spaces,
  capture(name = "captured_request_cookie", non_spaces),
  spaces,
  capture(name = "captured_response_cookie", non_spaces),
  spaces,
  capture(name = "termination_state", non_spaces),
  spaces,
  capture(name = "actconn", digits),
  "/",
  capture(name = "feconn", digits),
  "/",
  capture(name = "beconn", digits),
  "/",
  capture(name = "srv_conn", digits),
  "/",
  capture(name = "retries", some_of("+", digit)),
  spaces,
  capture(name = "srv_queue", digits),
  "/",
  capture(name = "backend_queue", digits),
  spaces,
  "{",
    capture(name = "captured_request_headers", except_any_of("}")),
  "}",
  spaces,
  "{",
    capture(name = "captured_response_headers", except_any_of("}")),
  "}",
  spaces,
  double_quote,
    capture(name = "http_request", non_quotes),
  double_quote)

re_matches(x, re)

#' ### <http://stackoverflow.com/questions/27422350/extract-character-preceding-first-dot-in-a-string>
#' Using [rex](http://cran.r-project.org/web/packages/rex/index.html) may make this type of task a little simpler.

my.data <- read.table(text = '
     my.string  state
     .........    A
     1........    B
     112......    C
     11111....    D
     1111113..    E
     111111111    F
     111111111    G
', header = TRUE, stringsAsFactors = FALSE)

library(rex)

re_matches(my.data$my.string,
  rex(capture(except(".")), "."))$'1'

#' ### <http://stackoverflow.com/questions/27410736>
#' Using [rex](http://cran.r-project.org/web/packages/rex/index.html) may make this type of task a little simpler.
string <- "Shakira - Wolf - 02.Hips don't lie.mp3"

library(rex)
re_matches(string,
  rex(capture(zero_or_more(any, type='lazy')), spaces, "-"))$'1'

#' ### <http://stackoverflow.com/questions/27400286/>
#' Using [rex](http://cran.r-project.org/web/packages/rex/index.html) may make this type of task a little simpler.

string <- "I t is tim e to g o" 
library(rex)
re_substitutes(string, rex(
    space %if_next_is%
      list(
        list(non_space, space, at_least(non_space, 2)) %or%
        list(non_space, end)
      )
    ), "", global = TRUE)

#' ### <http://stackoverflow.com/questions/27553126>
#' Using [rex](http://cran.r-project.org/web/packages/rex/index.html) may make this type of task a little simpler.

string <- "01:04:43.064 [12439] <2> xyz
01:04:43.067 [12439] <2> a lmn
01:04:43.068 [12439] <4> j klm
x_times_wait to <3000>
01:04:43.068 [12439] <4> j klm
enter_object <5000> main k"

library(rex)

timestamp <- rex(n(digit, 2), ":", n(digit, 2), ":", n(digit, 2), ".", n(digit, 3))

re <- rex(timestamp, space,
          "[", digits, "]", space,
          "<", digits, ">", space,
          capture(anything))

re_matches(string, re, global = TRUE)

shortcut <- function(...) {
  data <- list(...)
  if (length(data) == 1L) {
    data <- data[[1L]]
  }
  structure(data, class = "shortcut")
}

#' Single shortcuts
#'
#' Each of these shortcuts has both a plural (-s) and inverse (non_) form.
single_shortcuts <- shortcut(

  ## Character class shortcuts
  alnum = character_class("[:alnum:]"),
  alpha = character_class("[:alpha:]"),
  letter = character_class("[:alpha:]"),
  blank = character_class("[:blank:]"),
  cntrl = character_class("[:cntrl:]"),
  digit = character_class("[:digit:]"),
  number = character_class("[:digit:]"),
  graph = character_class("[:graph:]"),
  lower = character_class("[:lower:]"),
  print = character_class("[:print:]"),
  punct = character_class("[:punct:]"),
  space = character_class("[:space:]"),
  upper = character_class("[:upper:]"),
  xdigit = character_class("[:xdigit:]"),
  newline = regex("\\R"),

  single_quote = character_class("'"),
  double_quote = character_class("\""),
  quote = character_class("'\"")
)

basic_shortcuts <- shortcut(

  dot = escape("."),
  any = any_char <- regex("."),
  something = regex(".+"),
  anything = regex(".*"),

  start = regex("^"),
  end = regex("$"),

  boundary = regex("\\b"),
  non_boundary = regex("\\B")
)

inverse <- function(x) {
  x[] <- lapply(x, function(xx) {
    val <- paste0("^", xx); class(val) <- class(xx)
    val
  })
  names(x) <- paste0("non_", names(x))
  x
}

plural <- function(x) {
  x[] <- lapply(x, function(xx) {
    val <- paste0(escape(xx), "+"); class(val) <- "regex"
    val
  })
  names(x) <- paste0(names(x), "s")
  x
}

multiple <- function(x) {
  x[] <- lapply(x, function(xx) {
    val <- paste0(escape(xx), "*"); class(val) <- "regex"
    val
  })
  names(x) <- paste0("any_", names(x), "s")
  x
}

#' Shortcuts
#'
#' Commonly used character classes and regular expressions.  These shortcuts
#' are substituted inside \code{rex} calls.
#'
#' \code{names(shortcuts)} will give you the full list of available shortcuts.
#' @export
#' @family rex

shortcuts <- shortcut(c(
  basic_shortcuts,
  single_shortcuts,
  plural(single_shortcuts),
  multiple(single_shortcuts),
  inverse(single_shortcuts),
  plural(inverse(single_shortcuts)),
  multiple(inverse(single_shortcuts))
))

default_data_format.shortcut <- function(x) {
  build_rd <- get("build_rd", envir = asNamespace("roxygen2"))
  rd <- get("rd", envir = asNamespace("roxygen2"))

  build_rd(rd("\\preformatted{"), paste0(names(x), " - ", x, collapse = "\n"), rd("}"))
}

register_object(shortcuts)

context("helper_functions")

test_that("inverse negates the shortcut", {
  test_shortcuts <- shortcuts[c("single_quote", "letter")]

  expect_equal(
    list(
      non_single_quote = structure("^'", class = c("character_class", "regex")),
      non_letter = structure("^[:alpha:]", class = c("character_class", "regex"))
    ), inverse(test_shortcuts))
})
test_that("plural makes shortcut match one or more", {
  test_shortcuts <- shortcuts[c("single_quote", "letter")]

  expect_equal(
    list(
      single_quotes = structure("[']+", class = c("regex")),
      letters = structure("[[:alpha:]]+", class = c("regex"))
    ), plural(test_shortcuts))
})
test_that("multiple makes shortcut match zero or more", {
  test_shortcuts <- shortcuts[c("single_quote", "letter")]

  expect_equal(
    list(
      any_single_quotes = structure("[']*", class = c("regex")),
      any_letters = structure("[[:alpha:]]*", class = c("regex"))
    ), multiple(test_shortcuts))
})

#compdef rex
# zsh completions for 'rex'
# initially generated with http://github.com/RobSis/zsh-completion-generator
# complete groups, environments, hosts and tasks (default)

local curcontext="$curcontext" state state_descr line context ret=1

typeset -A opt_args
_hostgroups(){
	groups=(${(f)"$(rex -Ty 2>/dev/null| perl -MYAML -e 'my $groups = Load(join "", <>)->{groups}; print "$_\n" for keys %$groups;')"})
	_wanted hostgroups expl "available host groups" compadd -a groups
}

_envs(){
	envs=(${(f)"$(rex -Ty 2>/dev/null| perl -MYAML -e 'my $envs = Load(join "", <>)->{envs}; print "$_\n" for @$envs;')"})
	_wanted environments expl "available environments" compadd -a envs
}

# return hosts managed by rex or any other host availabe via zsh's _hosts function
_rex_hosts() {
	rexhosts=(${(f)"$(rex -Ty 2>/dev/null| perl -MYAML -MList::MoreUtils=uniq -E 'my $groups = Load(join "", <>)->{groups}; say $_->{name} for uniq sort map { @{ $groups->{$_} } } keys %$groups')"})
	_wanted hosts expl "rex managed hosts" compadd -a rexhosts || _hosts
}

local arguments
arguments=(
	'-b[run batch]'
	'-e[run the given code fragment]'
	'-E[execute a task on the given environment]:environments:_envs'
	'-G[|-g  Execute a task on the given server groups]:hosts group:_hostgroups'
	'-H[execute a task on the given hosts (space delimited)]:host:_rex_hosts'
	'-z[execute a task on hosts from this commands output]'
	'-K[public key file for the ssh connection]'
	'-P[private key file for the ssh connection]'
	'-p[password for the ssh connection]'
	'-u[username for the ssh connection]'
	'-d[show debug output]'
	'-ddd[show more debug output (includes profiling output)]'
	'-m[monochrome output: no colors]'
	'-o[output format]'
	'-q[quiet mode: no log output]'
	'-qw[quiet mode: only output warnings and errors]'
	'-Q[really quiet: output nothing]'
	'-T[list tasks]'
	'-Ta[List all tasks, including hidden]'
	'-Tm[list tasks in machine-readable format]'
	'-Tv[list tasks verbosely]'
	'-Ty[list tasks in YAML format]'
	'-c[turn cache ON]'
	'-C[turn cache OFF]'
	'-f[use this file instead of Rexfile]:filename:_files'
	'-F[force: disregard lock file]'
	'-h[display this help message]'
	'-M[load this module instead of Rexfile]'
	'-O[pass additional options, like CMDB path]'
	'-s[use sudo for every command]'
	'-S[password for sudo]'
	'-t[number of threads to use (aka parallelism param)]'
	'-v[display (R)?ex version]'
	'*:options:->vary'
)

_arguments -C -s -A "*" $arguments  && ret=0

case "$state" in
	vary)
		local optsfile
		optsfile='Rexfile'
		if [[ -e $optsfile ]]; then
			tasks=(${(f)"$(rex -Ty 2>/dev/null| perl -MYAML -E 'my $tasks = Load(join "", <>)->{tasks}; say $_ for @$tasks;')"})
			_wanted tasks expl "available tasks" compadd -a tasks
		fi
		;;
esac

return $ret

# bash completion for rex

_rex()
{
    local cur prev

    COMPREPLY=()
    _get_comp_words_by_ref -n : cur prev

    if [[ "$cur" == -* ]]; then
        COMPREPLY=( $(compgen -W '$(_parse_help "$1" -h)' -- "$cur") )
        return
    fi

    case "$prev" in
        -f)
            _filedir
            ;;
        -H)
            if [ -f Rexfile ]; then
                hosts=( $(rex -Ty 2>/dev/null | perl -MYAML -MList::MoreUtils=uniq -E 'my $groups = Load(join "", <>)->{groups}; say $_->{name} for uniq sort map { @{ $groups->{$_} } } keys %$groups') )
                COMPREPLY=( $( compgen -W '${hosts[@]}' -- "$cur" ) ) || _known_hosts_real -a "$cur"
            fi
            ;;
        -E)
            if [ -f Rexfile ]; then
                envs=( $(rex -Ty 2>/dev/null | perl -MYAML -e 'my $envs = Load(join "", <>)->{envs}; print "$_\n" for @$envs;') )
                COMPREPLY=( $( compgen -W '${envs[@]}' -- "$cur" ) )
            fi
            ;;
        -G)
            if [ -f Rexfile ]; then
                groups=( $(rex -Ty 2>/dev/null | perl -MYAML -e 'my $groups = Load(join "", <>)->{groups}; print "$_\n" for keys %$groups;') )
                COMPREPLY=( $( compgen -W '${groups[@]}' -- "$cur" ) )
            fi
            ;;

        *)
            if [ -f Rexfile ]; then
                tasks=( $(rex -Ty 2>/dev/null | perl -MYAML -E 'my $tasks = Load(join "", <>)->{tasks}; say $_ for @$tasks;') )
                COMPREPLY=( $( compgen -W '${tasks[@]}' -- "$cur" ) )
            fi
            ;;
    esac

    _rex_fix_colon_reply
    return 0

} &&
complete -F _rex rex

_rex_fix_colon_reply()
{
    local colprefs i
    colprefs=${cur%"${cur##*:}"}
    i=${#COMPREPLY[*]}
    while [ $((--i)) -ge 0 ]; do
        COMPREPLY[$i]=${COMPREPLY[$i]#"$colprefs"}
    done
}

# Local variables:
# mode: shell-script
# indent-tabs-mode: nil
# End:
# ex: ts=4 sw=4 et filetype=sh

---
title: "URL Validation"
author: "Jim Hester"
date: "`r Sys.Date()`"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{URL Validation}
  %\VignetteEngine{knitr::rmarkdown}
  \usepackage[utf8]{inputenc}
---

Consider the task of correctly [validating a URL](https://mathiasbynens.be/demo/url-regex).
From that page two conclusions can be made.

1. Validating URLs require complex regular expressions.
2. Creating a correct regular expression is hard! (only 1 out of 13 regexs were valid for all cases).

Because of this one may be tempted to simply copy the best regex you can find ([gist](https://gist.github.com/dperini/729294)).

The problem with this is that while you can copy it now, what happens later when you find a case that is not handled correctly?  Can you correctly interpret and modify this?
```{r url_parsing_stock, eval=F}
"^(?:(?:http(?:s)?|ftp)://)(?:\\S+(?::(?:\\S)*)?@)?(?:(?:[a-z0-9\u00a1-\uffff](?:-)*)*(?:[a-z0-9\u00a1-\uffff])+)(?:\\.(?:[a-z0-9\u00a1-\uffff](?:-)*)*(?:[a-z0-9\u00a1-\uffff])+)*(?:\\.(?:[a-z0-9\u00a1-\uffff]){2,})(?::(?:\\d){2,5})?(?:/(?:\\S)*)?$"
```

However if you re-create the regex with `rex` it is much easier to understand and modify later if needed.
```{r url_parsing_url}
library(rex)
library(magrittr)

valid_chars <- rex(except_some_of(".", "/", " ", "-"))

re <- rex(
  start,

  # protocol identifier (optional) + //
  group(list("http", maybe("s")) %or% "ftp", "://"),

  # user:pass authentication (optional)
  maybe(non_spaces,
    maybe(":", zero_or_more(non_space)),
    "@"),

  #host name
  group(zero_or_more(valid_chars, zero_or_more("-")), one_or_more(valid_chars)),

  #domain name
  zero_or_more(".", zero_or_more(valid_chars, zero_or_more("-")), one_or_more(valid_chars)),

  #TLD identifier
  group(".", valid_chars %>% at_least(2)),

  # server port number (optional)
  maybe(":", digit %>% between(2, 5)),

  # resource path (optional)
  maybe("/", non_space %>% zero_or_more()),

  end
)
```

We can then validate that it correctly identifies both good and bad URLs. (_IP address validation removed_)

```{r url_parsing_validate}
good <- c("http://foo.com/blah_blah",
  "http://foo.com/blah_blah/",
  "http://foo.com/blah_blah_(wikipedia)",
  "http://foo.com/blah_blah_(wikipedia)_(again)",
  "http://www.example.com/wpstyle/?p=364",
  "https://www.example.com/foo/?bar=baz&inga=42&quux",
  "http://✪df.ws/123",
  "http://userid:password@example.com:8080",
  "http://userid:password@example.com:8080/",
  "http://userid@example.com",
  "http://userid@example.com/",
  "http://userid@example.com:8080",
  "http://userid@example.com:8080/",
  "http://userid:password@example.com",
  "http://userid:password@example.com/",
  "http://➡.ws/䨹",
  "http://⌘.ws",
  "http://⌘.ws/",
  "http://foo.com/blah_(wikipedia)#cite-1",
  "http://foo.com/blah_(wikipedia)_blah#cite-1",
  "http://foo.com/unicode_(✪)_in_parens",
  "http://foo.com/(something)?after=parens",
  "http://☺.damowmow.com/",
  "http://code.google.com/events/#&product=browser",
  "http://j.mp",
  "ftp://foo.bar/baz",
  "http://foo.bar/?q=Test%20URL-encoded%20stuff",
  "http://مثال.إختبار",
  "http://例子.测试",
  "http://-.~_!$&'()*+,;=:%40:80%2f::::::@example.com",
  "http://1337.net",
  "http://a.b-c.de",
  "http://223.255.255.254")

bad <- c(
  "http://",
  "http://.",
  "http://..",
  "http://../",
  "http://?",
  "http://??",
  "http://??/",
  "http://#",
  "http://##",
  "http://##/",
  "http://foo.bar?q=Spaces should be encoded",
  "//",
  "//a",
  "///a",
  "///",
  "http:///a",
  "foo.com",
  "rdar://1234",
  "h://test",
  "http:// shouldfail.com",
  ":// should fail",
  "http://foo.bar/foo(bar)baz quux",
  "ftps://foo.bar/",
  "http://-error-.invalid/",
  "http://-a.b.co",
  "http://a.b-.co",
  "http://0.0.0.0",
  "http://3628126748",
  "http://.www.foo.bar/",
  "http://www.foo.bar./",
  "http://.www.foo.bar./")

all(grepl(re, good) == TRUE)

all(grepl(re, bad) == FALSE)
```

You can now see the power and expressiveness of building regular expressions with `rex`!

# Contributing guide

Thank you for considering to contribute to Rex! Volunteers like you make this
free and open source project possible.

We consider it important to ensure everyone involved can make an efficient use
of the time and effort they choose to invest. This guide aims to streamline
communication around contributions by documenting the most frequently discussed
aspects.

While most sections focus on the code contributions we typically receive, we
encourage and welcome other types of contributions too. For example:

- improving documentation
- testing with different setups
- donating (access to) hardware
- submitting and triaging bug reports
- sharing your use cases and experience with Rex
- spreading the word via blogs, talks, and social media
- starring the project on [GitHub](https://github.com/RexOps/Rex)
- adding Rex to your favorites on [MetaCPAN](https://metacpan.org/pod/Rex)
- listing Rex as part of your stack on [StackShare](https://stackshare.io/rex)

## Project guidelines

We took many decisions around Rex since the first commit in 2010. We found some
of the ideas coming up more often than others, and decided to share them here
(in no particular order).

### Getting help and support

If you need help with questions like “How do I…?” and “Why does it…?”, please
start a discussion via one of our
[support](https://www.rexify.org/support/index.html) channels.

We list options there for both community and commercial support.

### Discussions, issues, and pull requests

In short, we follow these steps in that order: discuss first, track next,
implement last.

1. Discuss: first we discuss the relevant details of any change proposal idea,
   and decide on next steps.
1. Track: when we decide to work on a change proposal, we open an issue to
   track progress, and design or refine potential solutions.
1. Implement: we open pull requests as a last step to propose and review a
   specific implementation of the given idea.

Please use the templates we provide to streamline our collaboration in typical
situations.

### Making exceptions

We may choose and accept skipping one or more of the steps in exceptional
cases, as long as the reasoning makes sense and gets well documented. When we
find ourselves repeating the same exception, we will document it here as a
standard case instead.

When you feel our workflow, guidelines, or templates does not fit your
situation, please consider that a strong indicator to [contact
us](https://www.rexify.org/support/index.html) early, and we’ll help figuring
out the best course of action.

### Rex core vs extending Rex

Strictly speaking, we consider the following the core competency of Rex:
execute commands and manage files, by defining tasks and orchestrating their
execution.

Rex gained lots of other capabilities over time, and historically many of them
landed in core as well. This includes the capability to extend Rex without
changing the core (or only minimally), for example when adding support to:

- new operating systems
- new shell types
- new virtualization methods
- new cloud providers

We strongly encourage to add such new capabilities via their own extension
modules outside the core. In case of questions or concerns, see also the
“Common scenarios” section, or [contact
us](https://www.rexify.org/support/index.html).

### Supported operating systems

We expect Rex to _run_ on operating systems wherever Perl can run. This
includes Linux, BSDs, Mac OS X, Windows, Solaris, and possibly others.

We also expect Rex to _manage_ endpoints running different operating systems.
This primarily means Unix-like systems for now.

We welcome patches and guides about how to support Rex both to run on and
manage even more operating systems.

In general, we aim to support Rex to run on and manage actively maintained
platforms only. See also the “Deprecation policy” section for more details.

### Supported Perl versions

Rex aims to run even on older Perl versions up to 10 years old. Currently this
means 5.14.4.

On top of the supported minimum version of Perl, we aim to support the latest
versions of all minor stable Perl 5 releases since then. That makes the full
list the following:

- 5.14.4
- 5.16.3
- 5.18.4
- 5.20.3
- 5.22.4
- 5.24.4
- 5.26.3
- 5.28.3
- 5.30.3
- 5.32.1
- 5.34.1
- 5.36.1
- 5.38.2
- 5.40.2

### Backwards compatibility

We aim to remain backwards compatible within major versions of Rex, as implied
by following [Dotted Semantic
Versioning](https://metacpan.org/pod/Version::Dotted::Semantic#Dotted-Semantic-Versioning-Specification).

To introduce new features while staying backwards compatible, Rex also uses
[feature flags](https://metacpan.org/pod/Rex#FEATURE-FLAGS). This makes it
possible to selectively opt-in to or opt-out from new features, depending on
the needs of the given use case.

Feature flags named after the minor releases of Rex (for example `1.4`) enable
the collection of preferred settings for the given Rex version.

In general, we strive to add new feature flags in pairs for opting in and out.

### Deprecation policy

We may drop features and code paths by following a planned deprecation
procedure. Adequate warnings and an ample transition period should precede any
such change.

If a supported operating system gets deprecated, we may try to keep it
supported while it doesn't cause any problems. As soon as a retired operating
system starts causing bugs, or gets in the way of progress, we consider it a
candidate to drop from support.

In other words, while we may choose to stay lazy to keep the support around, we
also don’t plan to put active effort into maintaining compatibility with
something which got abandoned by its own creators.

In case you depend on a deprecated feature, or must use or manage a retired
operating system, please consider contributing your solutions, or asking for
community or commercial [support](https://www.rexify.org/support/index.html)
for your use case.

### Code layout

We use the latest version of [Perl::Tidy](https://metacpan.org/pod/Perl::Tidy)
to format the Rex codebase according to the rules described in `.perltidyrc`.
We expect all contributions to follow the same formatting rules, and we also
actively test for it.

To avoid unnecessary “tidy only” commits in contributions, we recommend to
integrate a formatting step directly into your workflow. For example via a git
pre-commit hook, a shortcut in the editor, or as an automatic action upon
saving.

Importantly, we emphasize the importance of a consistent layout throughout the
codebase over any specific formatting rule. Since we include `.perltidyrc` in
the repo, consider it as subject of contributions too.

### Code quality

We use
[Test::Perl::Critic::Progressive](https://metacpan.org/pod/Test::Perl::Critic::Progressive)
to make sure every change follows the code quality rules in `.perlcriticrc`,
and no contribution introduces any violations accidentally.

The `dist.ini` file contains the list of perlcritic policy modules we use.

Since we include both `.perlcriticrc` and `dist.ini` in the repo, consider it
as subject of contributions too. In fact, we welcome improving both our
policies and the codebase in this regard.

### Tests should pass

Rex has two major test suites:

- the tests included with the code, exercising different modules and
  features
- the historical tests in the
  [RexOps/rex-build](https://github.com/RexOps/rex-build) repo, making sure Rex
  can manage actual VMs running different operating systems

In general, we expect contributions to include their own tests for the proposed
changes, while ensuring that previous tests still pass.

To make it easier to follow the guidelines of this document, we strive to have
corresponding tests for them as well.

### Have fun

We find it highly important to have fun while using Rex and contributing to it.
If you have a different experience, then please let us know, as that probably
indicates a bug somewhere.

## Development tools

We use [Git](https://www.git-scm.com) for version control of the source code,
and host Rex code in a [GitHub repository](https://github.com/RexOps/Rex).

We recommend to use a separate Perl environment for development, like the ones
you managed by [Perlbrew](https://perlbrew.pl/) or similar tools.

The code and examples use [cpanm](https://metacpan.org/pod/App::cpanminus) to
install modules from CPAN.

We also use [Dist::Zilla](https://metacpan.org/pod/Dist::Zilla) as an authoring
tool.

[Perltidy](https://metacpan.org/pod/distribution/Perl-Tidy/bin/perltidy) takes
care of maintaining a consistent source code formatting.

[Perlcritic](https://metacpan.org/pod/distribution/Perl-Critic/bin/perlcritic)
makes sure the codebase follows best practices and adheres to our minimum
quality requirements.

## Dependencies

After cloning the source code, run the following commands to install dependencies:

    dzil authordeps | cpanm
    dzil listdeps | cpanm

While some optional dependencies may not support your platform, use this
command to install these as well:

    dzil listdeps --suggests | cpanm

Install the remaining dependencies specific to the operating system, and Rex
itself with:

    dzil install

To only install the dependencies specific to the operating system, run one of
these commands:

- Windows: `cpanm Net::SSH2`
- non-Windows: `cpanm Net::OpenSSH Net::SFTP::Foreign IO::Pty`

When considering new dependencies, we prefer:

- Perl core modules
- CPAN modules already packaged by the operating systems distributing Rex
- CPAN modules supporting the same platforms Rex itself supports

We welcome and encourage contributions aimed at reducing the amount of Rex
dependencies, for example by:

- removing unused dependencies
- consolidating similar dependencies into a single one
- replacing external dependencies with Perl code modules

## Testing

Run the test suite included with the source code of Rex with `prove`:

    prove --lib --recurse t/

Extended, author, and release tests may need further dependencies. Install
these with:

    dzil listdeps --author --missing | cpanm

Then run those tests with:

    dzil test --all

As an important step before modifying the codebase, please run the progressive
perlcritic tests on the default branch:

    rm -f xt/author/.perlcritic-history
    AUTHOR_TESTING=1 prove --lib xt/author/critic-progressive.t

The first run should succeed, and records the current state. This serves as
baseline to compare later runs against. Please see the
[Test::Perl::Critic::Progressive
NOTES](https://metacpan.org/pod/Test::Perl::Critic::Progressive#NOTES) for more
details.

## Git workflow

We prefer receiving contributions by
[forking](https://docs.github.com/en/github/collaborating-with-issues-and-pull-requests/about-forks)
the repository on GitHub, and then sending [pull
requests](https://docs.github.com/en/github/collaborating-with-issues-and-pull-requests/about-pull-requests)
against the default branch of the repository. If you have to use a different
way, please [contact us](https://www.rexify.org/support/index.html) first to
discuss alternative options.

Please read and follow the “Discussions, issues, and pull requests” section
before investing effort in a pull request.

We recommend to use feature
[branches](https://docs.github.com/en/github/collaborating-with-issues-and-pull-requests/about-branches)
when working on contributions. This helps keeping the commits related to the
change together, while separating them from the main line of development.

We consider a single commit ideal when it:

- represents a single logical change ([atomic](https://en.wikipedia.org/wiki/Atomic_commit#Atomic_commit_convention))
- has a useful commit message
- has the expected test suite result

We expect commit messages to follow the guidelines collected in the [How to
Write a Git Commit Message](https://chris.beams.io/posts/git-commit/) article,
and we also test for most of those expectations.

We generally recommend to:

- follow a test-driven approach: add new breaking tests on a first commit, then
  change the code to fix them on a follow up commit, then refactoring on
  follow-up commits as needed
- simplify making larger changes: refactor on a separate commit first to make
  later changes simpler on smaller commits
- use atomic commits: use separate commits in a single pull request to separate
  the logical steps, and to help understanding the series of changes during
  review
- share your ideas and ask for feedback early: open [draft pull
  requests](https://docs.github.com/en/github/collaborating-with-issues-and-pull-requests/about-pull-requests#draft-pull-requests)
- keep your branch up-to-date:
  [rebase](https://docs.github.com/en/github/using-git/about-git-rebase) your
  feature branch on top of our default branch whenever we made new commits
  while you work
- keep a tidy history: use follow-up/cleanup commits on the same PR, and
  [squash or fixup related
  commits](https://docs.github.com/en/github/using-git/about-git-rebase)
  together in the feature branch, while keeping commits atomic (in other words,
  no need for “tidy only”, “fix typo”, and merge commits, while avoiding to
  squash everything into a single commit)
- use force push on pull request branches: work on your contributions until
  they pass automated tests and review, and keep the history tidy there to
  prepare for merging

## Common scenarios

### Add support to manage new operating systems

Allowing Rex to manage a new operating system requires the following steps:

1. Teach rex about how to detect the given operating system

    - add a way to `Rex::Hardware::Host::get_operating_system()` to detect the
      given operating system
    - add a new `is_myos()` function to `Rex::Commands::Gather`

1. Let Rex choose the proper package and service management modules for the
   given operating system

    - add support in `Rex::Service` and `Rex::Pkg`

1. Add new service and package management modules specific to the given
   operating system

    - add `Rex::Service::MyOS`
    - add `Rex::Pkg::MyOS`

While the first two steps must currently go into Rex core, please consider
publishing the `Rex::Service::MyOS` and `Rex::Pkg::MyOS` modules as a separate
distribution.

### Add support for new virtualization methods

Follow these steps to add support for a new virtualization method called `MyVirt`:

- create the top-level `Rex::Virtualization::MyVirt` module which includes the
  constructor, and the documentation
- create submodules for each virtualization command, e.g.
  `Rex::Virtualization::MyVirt::info`
- implement the logic of the given command as the `execute` method

Please consider publishing the `Rex::Virtualization::MyVirt` module as a
separate distribution.

## Contribute to this guide

If you think some of the information in this document got outdated, lacks
clarity, or has bugs, please propose changes to it too.

## Useful resources

- [Rex website](https://www.rexify.org)
- [MetaCPAN](https://metacpan.org/pod/Rex)
- [GitHub](https://github.com/RexOps/Rex)
- [Discussions](https://github.com/RexOps/Rex/discussions)
- [Issue tracker](https://github.com/RexOps/Rex/issues)
- [Google Groups](https://groups.google.com/forum/#!forum/rex-users)
- [StackShare](https://stackshare.io/rex)
- [Matrix](https://matrix.to/#/#rexops:matrix.org)
- [IRC](https://webchat.oftc.net/?channels=rexops)

---
title: "Server Log Parsing"
author: "Jim Hester"
date: "`r Sys.Date()`"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Server Log Parsing}
  %\VignetteEngine{knitr::rmarkdown}
  \usepackage[utf8]{inputenc}
---

Parsing server log files is a common task in server administration.
[1](https://link.springer.com/article/10.1007/BF03325089),[2](https://stackoverflow.com/search?q=%22Apache+log%22)
Historically R would not be well suited to this and it would be better
performed using a scripting language such as perl. Rex, however, makes this
easy to do and allows you to perform both the data cleaning and analysis in R!

Common server logs consist of space separated fields.

> 198.214.42.14 - - [21/Jul/1995:14:31:46 -0400] "GET /images/ HTTP/1.0" 200 17688

> lahal.ksc.nasa.gov - - [24/Jul/1995:12:42:40 -0400] "GET /images/USA-logosmall.gif HTTP/1.0" 200 234

The logs used in this vignette come from two months of all HTTP requests
to the NASA Kennedy Space Center WWW server in Florida and are freely available
for use. [3](https://web.archive.org/web/20181003084945/http://ita.ee.lbl.gov/html/contrib/NASA-HTTP.html)

```{r include = FALSE}
library(rex)
library(dplyr)
library(knitr)
library(ggplot2)
library(magrittr)
```

```{r show.warnings=FALSE}
parsed <- scan("NASA.txt", what = "character", sep = "\n") %>%
  re_matches(
    rex(

      # Get the time of the request
      "[",
        capture(name = "time",
          except_any_of("]")
        ),
      "]",

      space, double_quote, "GET", space,

      # Get the filetype of the request if requesting a file
      maybe(
        non_spaces, ".",
        capture(name = "filetype",
          except_some_of(space, ".", "?", double_quote)
        )
      )
    )
  ) %>%
  mutate(filetype = tolower(filetype),
         time = as.POSIXct(time, format="%d/%b/%Y:%H:%M:%S %z"))
```

This gives us a nicely formatted data frame of the time and filetypes of the requests.
```{r echo = FALSE}
kable(head(parsed, n = 10))
```

We can also easily generate a histogram of the filetypes, or a plot of requests over time.
```{r FALSE, fig.show='hold', warning = FALSE, message = FALSE}
ggplot(na.omit(parsed)) + stat_count(aes(x=filetype))
ggplot(na.omit(parsed)) + geom_histogram(aes(x=time)) + ggtitle("Requests over time")
```

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/capture.R
\name{capture}
\alias{capture}
\alias{.}
\alias{capture_group}
\title{Create a capture group}
\usage{
capture(..., name = NULL)

capture_group(name)
}
\arguments{
\item{...}{\code{\link{shortcuts}}, R variables, text, or other \pkg{rex}
functions.}

\item{name}{of the group.  Unnamed capture groups are numbers starting at 1
in the order they appear in the regular expression.  If two groups have the
same name, the leftmost group is the used in any reference.}
}
\description{
Used to save the matched value within the group for use later in the regular
expression or to extract the values captured.  Both named and unnamed groups
can later be referenced using \code{\link{capture_group}}.
}
\examples{

# Match paired quotation marks
re <- rex(
  # first quotation mark
  capture(quotes),

  # match all non-matching quotation marks
  zero_or_more(except(capture_group(1))),

  # end quotation mark (matches first)
  capture_group(1)
)

#named capture - don't match apples to oranges
re <- rex(
  capture(name = "fruit", or("apple", "orange")),
  "=",
  capture_group("fruit")
)
}
\seealso{
\code{\link{group}} for grouping without capturing.  Perl 5 Capture
Groups \url{http://perldoc.perl.org/perlre.html#Capture-groups}

Other rex: 
\code{\link{\%or\%}()},
\code{\link{character_class}()},
\code{\link{counts}},
\code{\link{group}()},
\code{\link{lookarounds}},
\code{\link{not}()},
\code{\link{rex}()},
\code{\link{shortcuts}},
\code{\link{wildcards}}
}
\concept{rex}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/match.R
\name{re_substitutes}
\alias{re_substitutes}
\alias{substitutes}
\alias{s}
\title{Substitute regular expressions in a string with another string.}
\usage{
re_substitutes(data, pattern, replacement, global = FALSE, options = NULL, ...)
}
\arguments{
\item{data}{character vector to substitute}

\item{pattern}{regular expression to match}

\item{replacement}{replacement text to use}

\item{global}{substitute all occurrences}

\item{options}{option flags}

\item{...}{options passed to sub or gsub}
}
\description{
Substitute regular expressions in a string with another string.
}
\examples{
string <- c("this is a Test", "string")
re_substitutes(string, "test", "not a test", options = "insensitive")
re_substitutes(string, "i", "x", global = TRUE)
re_substitutes(string, "(test)", "not a \\\\1", options = "insensitive")
}
\seealso{
\code{\link{regexp}} Section "Perl-like Regular Expressions" for a
discussion of the supported options
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rex-mode.R
\name{rex_mode}
\alias{rex_mode}
\title{Toggles \pkg{rex} mode.}
\usage{
rex_mode()
}
\description{
While within rex mode, functions used within the \code{\link{rex}} function
are attached, so one can get e.g. auto-completion within editors.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/shortcuts.R
\docType{data}
\name{shortcuts}
\alias{shortcuts}
\title{Shortcuts}
\format{
An object of class \code{shortcut} of length 116.
}
\usage{
shortcuts
}
\description{
Commonly used character classes and regular expressions.  These shortcuts
are substituted inside \code{rex} calls.
}
\details{
\code{names(shortcuts)} will give you the full list of available shortcuts.
}
\seealso{
Other rex: 
\code{\link{\%or\%}()},
\code{\link{capture}()},
\code{\link{character_class}()},
\code{\link{counts}},
\code{\link{group}()},
\code{\link{lookarounds}},
\code{\link{not}()},
\code{\link{rex}()},
\code{\link{wildcards}}
}
\concept{rex}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rex.R
\name{as.regex}
\alias{as.regex}
\alias{as.regex.default}
\title{Coerce objects to a \code{\link{regex}}.}
\usage{
as.regex(x, ...)

\method{as.regex}{default}(x, ...)
}
\arguments{
\item{x}{Object to coerce to \code{\link{regex}}.}

\item{...}{further arguments passed to methods.}
}
\description{
Coerce objects to a \code{\link{regex}}.
}
\section{Methods (by class)}{
\itemize{
\item \code{default}: Simply escape the Object.
}}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/wildcards.R
\name{wildcards}
\alias{wildcards}
\alias{zero_or_more}
\alias{one_or_more}
\alias{maybe}
\alias{zero_or_one}
\title{Wildcards}
\usage{
zero_or_more(..., type = c("greedy", "lazy", "possessive"))

one_or_more(..., type = c("greedy", "lazy", "possessive"))

maybe(..., type = c("greedy", "lazy", "possessive"))
}
\arguments{
\item{...}{\code{\link{shortcuts}}, R variables, text, or other \pkg{rex}
functions.}

\item{type}{the type of match to perform.

There are three match types
\enumerate{
  \item \code{greedy}: match the longest string.  This is the default matching type.
  \item \code{lazy}: match the shortest string.  This matches the shortest string from the same anchor point, not necessarily the shortest global string.
  \item \code{possessive}: match and don't allow backtracking
}}
}
\description{
Wildcards
}
\section{Functions}{
\itemize{
\item \code{zero_or_more}: match \code{...} zero or more times.

\item \code{one_or_more}: match \code{...} one or more times.

\item \code{maybe}: match \code{...} zero or one times.
}}

\seealso{
Other rex: 
\code{\link{\%or\%}()},
\code{\link{capture}()},
\code{\link{character_class}()},
\code{\link{counts}},
\code{\link{group}()},
\code{\link{lookarounds}},
\code{\link{not}()},
\code{\link{rex}()},
\code{\link{shortcuts}}
}
\concept{rex}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/wildcards.R
\name{not}
\alias{not}
\title{Do not match}
\usage{
not(..., type = c("greedy", "lazy", "possessive"))
}
\arguments{
\item{...}{\code{\link{shortcuts}}, R variables, text, or other \pkg{rex}
functions.}

\item{type}{the type of match to perform.

There are three match types
\enumerate{
  \item \code{greedy}: match the longest string.  This is the default matching type.
  \item \code{lazy}: match the shortest string.  This matches the shortest string from the same anchor point, not necessarily the shortest global string.
  \item \code{possessive}: match and don't allow backtracking
}}
}
\description{
Do not match
}
\seealso{
Other rex: 
\code{\link{\%or\%}()},
\code{\link{capture}()},
\code{\link{character_class}()},
\code{\link{counts}},
\code{\link{group}()},
\code{\link{lookarounds}},
\code{\link{rex}()},
\code{\link{shortcuts}},
\code{\link{wildcards}}
}
\concept{rex}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rex.R
\name{register_shortcuts}
\alias{register_shortcuts}
\title{Register the Rex shortcuts}
\usage{
register_shortcuts(pkg_name)
}
\arguments{
\item{pkg_name}{the package to register the shortcuts in}
}
\description{
If you are using rex in another package you need to call this function to
register all of the rex shortcuts so that spurious NOTEs about global
variables being generated during R CMD check.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/capture.R
\name{group}
\alias{group}
\title{Create a grouped expression}
\usage{
group(...)
}
\arguments{
\item{...}{\code{\link{shortcuts}}, R variables, text, or other \pkg{rex}
functions.}
}
\description{
This is similar to \code{\link{capture}} except that it does not store the
value of the group.  Best used when you want to combine several parts
together and do not reference or extract the grouped value later.
}
\seealso{
\code{\link{capture}} for grouping with capturing.  Perl 5 Extended
Patterns \url{http://perldoc.perl.org/perlre.html#Extended-Patterns}

Other rex: 
\code{\link{\%or\%}()},
\code{\link{capture}()},
\code{\link{character_class}()},
\code{\link{counts}},
\code{\link{lookarounds}},
\code{\link{not}()},
\code{\link{rex}()},
\code{\link{shortcuts}},
\code{\link{wildcards}}
}
\concept{rex}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/counts.R
\name{counts}
\alias{counts}
\alias{n_times}
\alias{n}
\alias{between}
\alias{at_least}
\alias{at_most}
\title{Counts}
\usage{
n_times(x, n, type = c("greedy", "lazy", "possessive"))

between(x, low, high, type = c("greedy", "lazy", "possessive"))

at_least(x, n, type = c("greedy", "lazy", "possessive"))

at_most(x, n, type = c("greedy", "lazy", "possessive"))
}
\arguments{
\item{x}{A regex pattern.}

\item{n}{An integer number}

\item{type}{the type of match to perform.

There are three match types
\enumerate{
  \item \code{greedy}: match the longest string.  This is the default matching type.
  \item \code{lazy}: match the shortest string.  This matches the shortest string from the same anchor point, not necessarily the shortest global string.
  \item \code{possessive}: match and don't allow backtracking
}}

\item{low}{An integer number for the lower limit.}

\item{high}{An integer number for the upper limit.}
}
\description{
Functions to restrict a regex to a specific number
}
\section{Functions}{
\itemize{
\item \code{n_times}: \code{x} must occur exactly \code{n} times.

\item \code{between}: \code{x} must occur between \code{low} and \code{high} times.

\item \code{at_least}: \code{x} must occur at least \code{n} times.

\item \code{at_most}: \code{x} must occur at most \code{n} times.
}}

\seealso{
Other rex: 
\code{\link{\%or\%}()},
\code{\link{capture}()},
\code{\link{character_class}()},
\code{\link{group}()},
\code{\link{lookarounds}},
\code{\link{not}()},
\code{\link{rex}()},
\code{\link{shortcuts}},
\code{\link{wildcards}}
}
\concept{rex}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/escape.R
\name{character_class_escape}
\alias{character_class_escape}
\alias{character_class_escape.regex}
\alias{character_class_escape.character_class}
\alias{character_class_escape.character}
\alias{character_class_escape.list}
\alias{character_class_escape.default}
\title{Character class escapes}
\usage{
character_class_escape(x)

\method{character_class_escape}{regex}(x)

\method{character_class_escape}{character_class}(x)

\method{character_class_escape}{character}(x)

\method{character_class_escape}{list}(x)

\method{character_class_escape}{default}(x)
}
\arguments{
\item{x}{Object to escape.}
}
\description{
Character class escapes
}
\section{Methods (by class)}{
\itemize{
\item \code{regex}: objects are passed through unchanged.

\item \code{character_class}: objects are passed through unchanged.

\item \code{character}: objects properly escaped for character classes.

\item \code{list}: call \code{character_class_escape} on all elements of the list.

\item \code{default}: coerce to \code{character} and \code{character_class_escape}.
}}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/match.R
\name{re_matches}
\alias{re_matches}
\alias{matches}
\alias{m}
\title{Match function}
\usage{
re_matches(
  data,
  pattern,
  global = FALSE,
  options = NULL,
  locations = FALSE,
  ...
)
}
\arguments{
\item{data}{character vector to match against}

\item{pattern}{regular expression to use for matching}

\item{global}{use global matching}

\item{options}{regular expression options}

\item{locations}{rather than returning the values of the matched (or
captured) string, return a \code{data.frame} of the match locations in the
string.}

\item{...}{options passed to regexpr or gregexpr}
}
\value{
if no captures, returns a logical vector the same length as the
input character vector specifying if the relevant value matched or not.  If
there are captures in the regular expression, returns a \code{data.frame} with a
column for each capture group.  If \code{global} is \code{TRUE}, returns a
list of \code{data.frame}s.
}
\description{
Match function
}
\examples{
string <- c("this is a", "test string")
re_matches(string, rex("test")) # FALSE FALSE

# named capture
re_matches(string, rex(capture(alphas, name = "first_word"), space,
  capture(alphas, name = "second_word")))
#   first_word second_word
# 1       this          is
# 2       test      string

# capture returns NA when it fails to match
re_matches(string, rex(capture("test")))
#      1
# 1 test
# 2 <NA>
}
\seealso{
\code{\link{regexp}} Section "Perl-like Regular Expressions" for a
discussion of the supported options
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/shortcuts.R
\docType{data}
\name{single_shortcuts}
\alias{single_shortcuts}
\title{Single shortcuts}
\format{
An object of class \code{shortcut} of length 18.
}
\usage{
single_shortcuts
}
\description{
Each of these shortcuts has both a plural (-s) and inverse (non_) form.
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rex.R
\name{as.character.regex}
\alias{as.character.regex}
\alias{print.regex}
\alias{regex}
\title{Regular Expression}
\usage{
\method{as.character}{regex}(x, ...)

\method{print}{regex}(x, ...)

regex(x, ...)
}
\arguments{
\item{x}{Object}

\item{...}{further arguments}
}
\description{
Specify an explicit regular expression.  This expression must already be
escaped.
}
\section{Methods (by generic)}{
\itemize{
\item \code{as.character}: coerce regex object to a character

\item \code{print}: Print regex object
}}

\seealso{
\code{\link{as.regex}} to coerce to a regex object.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/lookarounds.R
\name{lookarounds}
\alias{lookarounds}
\alias{\%if_next_is\%}
\alias{\%if_next_isnt\%}
\alias{\%if_prev_is\%}
\alias{\%if_prev_isnt\%}
\title{Lookarounds}
\usage{
x \%if_next_is\% y

x \%if_next_isnt\% y

x \%if_prev_is\% y

x \%if_prev_isnt\% y
}
\arguments{
\item{x}{A regex pattern.}

\item{y}{A regex pattern.}
}
\description{
Lookarounds
}
\details{
These functions provide an interface to perl lookarounds.

Special binary functions are used to infer an ordering, since often you
might wish to match a word / set of characters conditional on the start
and end of that word.

\itemize{
  \item \code{\%if_next_is\%}: \code{TRUE} if x follows y
  \item \code{\%if_next_isnt\%}: \code{TRUE} if x does not follow y
  \item \code{\%if_prev_is\%}: \code{TRUE} if y comes before x
  \item \code{\%if_prev_isnt\%}: \code{TRUE} if y does not come before x
}
}
\examples{
stopifnot(grepl(rex("crab" \%if_next_is\% "apple"), "crabapple", perl = TRUE))
stopifnot(grepl(rex("crab" \%if_prev_is\% "apple"), "applecrab", perl = TRUE))
stopifnot(grepl(rex(range("a", "e") \%if_next_isnt\% range("f", "g")),
  "ah", perl = TRUE))
stopifnot(grepl(rex(range("a", "e") \%if_next_is\% range("f", "i")),
  "ah", perl = TRUE))
}
\seealso{
Perl 5 Documentation \url{http://perldoc.perl.org/perlre.html#Extended-Patterns}

Other rex: 
\code{\link{\%or\%}()},
\code{\link{capture}()},
\code{\link{character_class}()},
\code{\link{counts}},
\code{\link{group}()},
\code{\link{not}()},
\code{\link{rex}()},
\code{\link{shortcuts}},
\code{\link{wildcards}}
}
\concept{rex}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/character_class.R
\name{character_class}
\alias{character_class}
\alias{one_of}
\alias{any_of}
\alias{some_of}
\alias{none_of}
\alias{except}
\alias{except_any_of}
\alias{except_some_of}
\alias{range}
\alias{:}
\alias{exclude_range}
\title{Create character classes}
\usage{
character_class(x)

one_of(...)

any_of(..., type = c("greedy", "lazy", "possessive"))

some_of(..., type = c("greedy", "lazy", "possessive"))

none_of(...)

except_any_of(..., type = c("greedy", "lazy", "possessive"))

except_some_of(..., type = c("greedy", "lazy", "possessive"))

range(start, end)

`:`(start, end)

exclude_range(start, end)
}
\arguments{
\item{x}{text to include in the character class (must be escaped manually)}

\item{...}{\code{\link{shortcuts}}, R variables, text, or other \pkg{rex}
functions.}

\item{type}{the type of match to perform.

There are three match types
\enumerate{
  \item \code{greedy}: match the longest string.  This is the default matching type.
  \item \code{lazy}: match the shortest string.  This matches the shortest string from the same anchor point, not necessarily the shortest global string.
  \item \code{possessive}: match and don't allow backtracking
}}

\item{start}{beginning of character class}

\item{end}{end of character class}
}
\description{
There are multiple ways you can define a character class.
}
\section{Functions}{
\itemize{
\item \code{character_class}: explicitly define a character class

\item \code{one_of}: matches one of the specified characters.

\item \code{any_of}: matches zero or more of the specified characters.

\item \code{some_of}: matches one or more of the specified characters.

\item \code{none_of}: matches anything but one of the specified characters.

\item \code{except_any_of}: matches zero or more of anything but the specified characters.

\item \code{except_some_of}: matches one or more of anything but the specified characters.

\item \code{range}: matches one of any of the characters in the range.

\item \code{:}: matches one of any of the characters in the range.

\item \code{exclude_range}: matches one of any of the characters except those in the range.
}}

\examples{
# grey = gray
re <- rex("gr", one_of("a", "e"), "y")
grepl(re, c("grey", "gray")) # TRUE TRUE

# Match non-vowels
re <- rex(none_of("a", "e", "i", "o", "u"))
# They can also be in the same string
re <- rex(none_of("aeiou"))
grepl(re, c("k", "l", "e")) # TRUE TRUE FALSE

# Match range
re <- rex(range("a", "e"))
grepl(re, c("b", "d", "f")) # TRUE TRUE FALSE

# Explicit creation
re <- rex(character_class("abcd\\\\["))
grepl(re, c("a", "d", "[", "]")) # TRUE TRUE TRUE FALSE
}
\seealso{
Other rex: 
\code{\link{\%or\%}()},
\code{\link{capture}()},
\code{\link{counts}},
\code{\link{group}()},
\code{\link{lookarounds}},
\code{\link{not}()},
\code{\link{rex}()},
\code{\link{shortcuts}},
\code{\link{wildcards}}
}
\concept{rex}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/escape.R
\name{escape}
\alias{escape}
\alias{escape.regex}
\alias{escape.character_class}
\alias{escape.character}
\alias{escape.default}
\alias{escape.list}
\title{Escape characters for a regex}
\usage{
escape(x)

\method{escape}{regex}(x)

\method{escape}{character_class}(x)

\method{escape}{character}(x)

\method{escape}{default}(x)

\method{escape}{list}(x)
}
\arguments{
\item{x}{Object to escape.}
}
\description{
Escape characters for a regex
}
\section{Methods (by class)}{
\itemize{
\item \code{regex}: Objects are simply passed through unchanged.

\item \code{character_class}: Objects are surrounded by braces.

\item \code{character}: Objects are properly escaped for regular expressions.

\item \code{default}: default escape coerces to character and escapes.

\item \code{list}: simply call escape on all elements of the list.
}}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/or.R
\name{\%or\%}
\alias{\%or\%}
\alias{or}
\title{Or}
\usage{
x \%or\% y

or(...)
}
\arguments{
\item{x}{A string.}

\item{y}{A string.}

\item{...}{\code{\link{shortcuts}}, R variables, text, or other \pkg{rex}
functions.}
}
\description{
The special binary function \code{\%or\%} can be used to specify a set
of optional matches.
}
\seealso{
Other rex: 
\code{\link{capture}()},
\code{\link{character_class}()},
\code{\link{counts}},
\code{\link{group}()},
\code{\link{lookarounds}},
\code{\link{not}()},
\code{\link{rex}()},
\code{\link{shortcuts}},
\code{\link{wildcards}}
}
\concept{rex}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rex.R
\name{rex}
\alias{rex}
\alias{rex_}
\title{Generate a regular expression.}
\usage{
rex(..., env = parent.frame())
}
\arguments{
\item{...}{\code{\link{shortcuts}}, R variables, text, or other \pkg{rex}
functions.}

\item{env}{environment to evaluate the rex expression in.}
}
\description{
Generate a regular expression.
}
\seealso{
Other rex: 
\code{\link{\%or\%}()},
\code{\link{capture}()},
\code{\link{character_class}()},
\code{\link{counts}},
\code{\link{group}()},
\code{\link{lookarounds}},
\code{\link{not}()},
\code{\link{shortcuts}},
\code{\link{wildcards}}
}
\concept{rex}
